#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## card.py
##
##  Created on: Sep 26, 2017
##      Author: Alexey S. Ignatiev
##      E-mail: aignatiev@ciencias.ulisboa.pt
##

"""
    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        EncType
        CardEnc
        ITotalizer

    ==================
    Module description
    ==================

    This module provides access to various *cardinality constraint* [1]_
    encodings to formulas in conjunctive normal form (CNF). These include
    pairwise [2]_, bitwise [2]_, ladder/regular [3]_ [4]_, sequential counters
    [5]_, sorting [6]_ and cardinality networks [7]_, totalizer [8]_, modulo
    totalizer [9]_, and modulo totalizer for :math:`k`-cardinality [10]_, as
    well as a *native* cardinality constraint representation supported by the
    `MiniCard solver <https://github.com/liffiton/minicard>`__.

    .. [1] Olivier Roussel, Vasco M. Manquinho. *Pseudo-Boolean and Cardinality
        Constraints*. Handbook of Satisfiability.  2009. pp. 695-733

    .. [2] Steven David Prestwich. *CNF Encodings*. Handbook of Satisfiability.
        2009. pp. 75-97

    .. [3] Carlos Ansótegui, Felip Manyà. *Mapping Problems with Finite-Domain
        Variables to Problems with Boolean Variables*. SAT (Selected Papers)
        2004. pp. 1-15

    .. [4] Ian P. Gent, Peter Nightingale. *A New Encoding of Alldifferent Into
        SAT*. In International workshop on modelling and reformulating
        constraint satisfaction problems 2004. pp. 95-110

    .. [5] Carsten Sinz. *Towards an Optimal CNF Encoding of Boolean
        Cardinality Constraints*. CP 2005. pp. 827-831

    .. [6] Kenneth E. Batcher. *Sorting Networks and Their Applications*.
        AFIPS Spring Joint Computing Conference 1968. pp. 307-314

    .. [7] Roberto Asin, Robert Nieuwenhuis, Albert Oliveras,
        Enric Rodriguez-Carbonell. *Cardinality Networks and Their
        Applications*. SAT 2009. pp. 167-180

    .. [8] Olivier Bailleux, Yacine Boufkhad. *Efficient CNF Encoding of
        Boolean Cardinality Constraints*. CP 2003. pp. 108-122

    .. [9] Toru Ogawa, Yangyang Liu, Ryuzo Hasegawa, Miyuki Koshimura,
        Hiroshi Fujita. *Modulo Based CNF Encoding of Cardinality Constraints
        and Its Application to MaxSAT Solvers*. ICTAI 2013. pp. 9-17

    .. [10] António Morgado, Alexey Ignatiev, Joao Marques-Silva. *MSCG: Robust
        Core-Guided MaxSAT Solving*. System Description. JSAT 2015. vol. 9,
        pp. 129-134

    A cardinality constraint is a constraint of the form:
    :math:`\sum_{i=1}^n{x_i}\leq k`. Cardinality constraints are ubiquitous in
    practical problem formulations. Note that the implementation of the
    pairwise, bitwise, and ladder encodings can only deal with AtMost1
    constraints, e.g. :math:`\sum_{i=1}^n{x_i}\leq 1`.

    Access to all cardinality encodings can be made through the main class of
    this module, which is :class:`.CardEnc`.

    Additionally, to the standard cardinality encodings that are basically
    "static" CNF formulas, the module is designed to able to construct
    *incremental* cardinality encodings, i.e. those that can be incrementally
    extended at a later stage. At this point only the *iterative totalizer*
    [11]_ encoding is supported. Iterative totalizer can be accessed with the
    use of the :class:`.ITotalizer` class.

    .. [11] Ruben Martins, Saurabh Joshi, Vasco M. Manquinho, Inês Lynce.
        *Incremental Cardinality Constraints for MaxSAT*. CP 2014. pp. 531-548

    ==============
    Module details
    ==============
"""

#
#==============================================================================
import math
from pysat.formula import CNF, CNFPlus, IDPool
from pysat._utils import MainThread
import pycard
import signal


#
#==============================================================================
class NoSuchEncodingError(Exception):
    """
        This exception is raised when creating an unknown an AtMostk, AtLeastK,
        or EqualK constraint encoding.
    """

    pass


#
#==============================================================================
class EncType(object):
    """
        This class represents a C-like ``enum`` type for choosing the
        cardinality encoding to use. The values denoting the encodings are:

        ::

            pairwise    = 0
            seqcounter  = 1
            sortnetwrk  = 2
            cardnetwrk  = 3
            bitwise     = 4
            ladder      = 5
            totalizer   = 6
            mtotalizer  = 7
            kmtotalizer = 8
            native      = 9

        The desired encoding can be selected either directly by its integer
        identifier, e.g. ``2``, or by its alphabetical name, e.g.
        ``EncType.sortnetwrk``.

        Note that while most of the encodings are produced as a list of
        clauses, the "native" encoding of `MiniCard
        <https://github.com/liffiton/minicard>`__ is managed as one clause.
        Given an AtMostK constraint :math:`\sum_{i=1}^n{x_i\leq k}`, the native
        encoding represents it as a pair ``[lits, k]``, where ``lits`` is a
        list of size ``n`` containing literals in the sum.
    """

    pairwise    = 0
    seqcounter  = 1
    sortnetwrk  = 2
    cardnetwrk  = 3
    bitwise     = 4
    ladder      = 5
    totalizer   = 6
    mtotalizer  = 7
    kmtotalizer = 8
    native      = 9  # native representation used by Minicard


#
#==============================================================================
class CardEnc(object):
    """
        This abstract class is responsible for the creation of cardinality
        constraints encoded to a CNF formula. The class has three *class
        methods* for creating AtMostK, AtLeastK, and EqualsK constraints. Given
        a list of literals, an integer bound and an encoding type, each of
        these methods returns an object of class :class:`pysat.formula.CNFPlus`
        representing the resulting CNF formula.

        Since the class is abstract, there is no need to create an object of
        it. Instead, the methods should be called directly as class methods,
        e.g. ``CardEnc.atmost(lits, bound)`` or ``CardEnc.equals(lits,
        bound)``. An example usage is the following:

        .. code-block:: python

            >>> from pysat.card import *
            >>> cnf = CardEnc.atmost(lits=[1, 2, 3], encoding=EncType.pairwise)
            >>> print(cnf.clauses)
            [[-1, -2], [-1, -3], [-2, -3]]
            >>> cnf = CardEnc.equals(lits=[1, 2, 3], encoding=EncType.pairwise)
            >>> print(cnf.clauses)
            [[1, 2, 3], [-1, -2], [-1, -3], [-2, -3]]
    """

    @classmethod
    def _update_vids(cls, cnf, vpool):
        """
            Update variable ids in the given formula and id pool.

            :param cnf: a list of literals in the sum.
            :param vpool: the value of bound :math:`k`.

            :type cnf: :class:`.formula.CNFPlus`
            :type vpool: :class:`.formula.IDPool`
        """

        top, vmap = vpool.top, {}  # current top and variable mapping

        # creating a new variable mapping, taking into
        # account variables marked as "occupied"
        while top < cnf.nv:
            top += 1
            vpool.top += 1

            while vpool._occupied and vpool.top >= vpool._occupied[0][0]:
                if vpool.top <= vpool._occupied[0][1] + 1:
                    vpool.top = vpool._occupied[0][1] + 1

                vpool._occupied.pop(0)

            vmap[top] = vpool.top

        # updating the clauses
        for cl in cnf.clauses:
            cl[:] = map(lambda l: int(math.copysign(vmap[abs(l)], l)) if abs(l) in vmap else l, cl)

        # updating the number of variables
        cnf.nv = vpool.top

    @classmethod
    def atmost(cls, lits, bound=1, top_id=None, vpool=None,
            encoding=EncType.seqcounter):
        """
            This method can be used for creating a CNF encoding of an AtMostK
            constraint, i.e. of :math:`\sum_{i=1}^{n}{x_i}\leq k`. The method
            shares the arguments and the return type with method
            :meth:`CardEnc.atleast`. Please, see it for details.
        """

        if encoding < 0 or encoding > 9:
            raise(NoSuchEncodingError(encoding))

        assert not top_id or not vpool, \
                'Use either a top id or a pool of variables but not both.'

        # we are going to return this formula
        ret = CNFPlus()

        # if the list of literals is empty, return empty formula
        if not lits:
            return ret

        # obtaining the top id from the variable pool
        if vpool:
            top_id = vpool.top

        if not top_id:
            top_id = max(map(lambda x: abs(x), lits))

        # MiniCard's native representation is handled separately
        if encoding == 9:
            ret.atmosts, ret.nv = [(lits, bound)], top_id
            return ret

        if MainThread.check() == True:
            # saving default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_DFL)

            res = pycard.encode_atmost(lits, bound, top_id, encoding, 1)

            # recovering default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, def_sigint_handler)
        else:
            res = pycard.encode_atmost(lits, bound, top_id, encoding, 0)

        if res:
            ret.clauses, ret.nv = res

        # updating vpool if necessary
        if vpool:
            if vpool._occupied and vpool.top <= vpool._occupied[0][0] <= ret.nv:
                cls._update_vids(ret, vpool)
            else:
                vpool.top = ret.nv - 1
                vpool._next()

        return ret

    @classmethod
    def atleast(cls, lits, bound=1, top_id=None, vpool=None,
            encoding=EncType.seqcounter):
        """
            This method can be used for creating a CNF encoding of an AtLeastK
            constraint, i.e. of :math:`\sum_{i=1}^{n}{x_i}\geq k`. The method
            takes 1 mandatory argument ``lits`` and 3 default arguments can be
            specified: ``bound``, ``top_id``, ``vpool``, and ``encoding``.

            :param lits: a list of literals in the sum.
            :param bound: the value of bound :math:`k`.
            :param top_id: top variable identifier used so far.
            :param vpool: variable pool for counting the number of variables.
            :param encoding: identifier of the encoding to use.

            :type lits: iterable(int)
            :type bound: int
            :type top_id: integer or None
            :type vpool: :class:`.IDPool`
            :type encoding: integer

            Parameter ``top_id`` serves to increase integer identifiers of
            auxiliary variables introduced during the encoding process. This
            is helpful when augmenting an existing CNF formula with the new
            cardinality encoding to make sure there is no collision between
            identifiers of the variables. If specified, the identifiers of the
            first auxiliary variable will be ``top_id+1``.

            Instead of ``top_id``, one may want to use a pool of variable
            identifiers ``vpool``, which is automatically updated during the
            method call. In many circumstances, this is more convenient than
            using ``top_id``. Also note that parameters ``top_id`` and
            ``vpool`` **cannot** be specified *simultaneusly*.

            The default value of ``encoding`` is :attr:`Enctype.seqcounter`.

            The method *translates* the AtLeast constraint into an AtMost
            constraint by *negating* the literals of ``lits``, creating a new
            bound :math:`n-k` and invoking :meth:`CardEnc.atmost` with the
            modified list of literals and the new bound.

            :raises CardEnc.NoSuchEncodingError: if encoding does not exist.

            :rtype: a :class:`.CNFPlus` object where the new \
            clauses (or the new native atmost constraint) are stored.
        """

        if encoding < 0 or encoding > 9:
            raise(NoSuchEncodingError(encoding))

        assert not top_id or not vpool, \
                'Use either a top id or a pool of variables but not both.'

        # we are going to return this formula
        ret = CNFPlus()

        # if the list of literals is empty, return empty formula
        if not lits:
            return ret

        # obtaining the top id from the variable pool
        if vpool:
            top_id = vpool.top

        if not top_id:
            top_id = max(map(lambda x: abs(x), lits))

        # Minicard's native representation is handled separately
        if encoding == 9:
            ret.atmosts, ret.nv = [([-l for l in lits], len(lits) - bound)], top_id
            return ret

        if MainThread.check() == True:
            # saving default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_DFL)

            res = pycard.encode_atleast(lits, bound, top_id, encoding, 1)

            # recovering default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, def_sigint_handler)
        else:
            res = pycard.encode_atleast(lits, bound, top_id, encoding, 0)

        if res:
            ret.clauses, ret.nv = res

        # updating vpool if necessary
        if vpool:
            if vpool._occupied and vpool.top <= vpool._occupied[0][0] <= ret.nv:
                cls._update_vids(ret, vpool)
            else:
                vpool.top = ret.nv - 1
                vpool._next()

        return ret

    @classmethod
    def equals(cls, lits, bound=1, top_id=None, vpool=None,
            encoding=EncType.seqcounter):
        """
            This method can be used for creating a CNF encoding of an EqualsK
            constraint, i.e. of :math:`\sum_{i=1}^{n}{x_i}= k`. The method
            makes consecutive calls of both :meth:`CardEnc.atleast` and
            :meth:`CardEnc.atmost`. It shares the arguments and the return type
            with method :meth:`CardEnc.atleast`. Please, see it for details.
        """

        if vpool:
            res1 = cls.atleast(lits, bound=bound, vpool=vpool, encoding=encoding)
            res2 = cls.atmost(lits, bound=bound, vpool=vpool, encoding=encoding)
        else:
            res1 = cls.atleast(lits, bound=bound, top_id=top_id, encoding=encoding)
            res2 = cls.atmost(lits, bound=bound, top_id=res1.nv, encoding=encoding)

        # merging together AtLeast and AtMost constraints
        res1.nv = max(res1.nv, res2.nv)
        res1.clauses.extend(res2.clauses)
        res1.atmosts.extend(res2.atmosts)

        return res1


#
#==============================================================================
class ITotalizer(object):
    """
        This class implements the iterative totalizer encoding [11]_. Note that
        :class:`ITotalizer` can be used only for creating AtMostK constraints.
        In contrast to class :class:`EncType`, this class is not abstract and
        its objects once created can be reused several times. The idea is that
        a *totalizer tree* can be extended, or the bound can be increased, as
        well as two totalizer trees can be merged into one.

        The constructor of the class object takes 3 default arguments.

        :param lits: a list of literals to sum.
        :param ubound: the largest potential bound to use.
        :param top_id: top variable identifier used so far.

        :type lits: iterable(int)
        :type ubound: int
        :type top_id: integer or None

        The encoding of the current tree can be accessed with the use of
        :class:`.CNF` variable stored as ``self.cnf``. Potential bounds **are
        not** imposed by default but can be added as unit clauses in the final
        CNF formula. The bounds are stored in the list of Boolean variables as
        ``self.rhs``. A concrete bound :math:`k` can be enforced by considering
        a unit clause ``-self.rhs[k]``. **Note** that ``-self.rhs[0]`` enforces
        all literals of the sum to be *false*.

        An :class:`ITotalizer` object should be deleted if it is not needed
        anymore.

        Possible usage of the class is shown below:

        .. code-block:: python

            >>> from pysat.card import ITotalizer
            >>> t = ITotalizer(lits=[1, 2, 3], ubound=1)
            >>> print(t.cnf.clauses)
            [[-2, 4], [-1, 4], [-1, -2, 5], [-4, 6], [-5, 7], [-3, 6], [-3, -4, 7]]
            >>> print(t.rhs)
            [6, 7]
            >>> t.delete()

        Alternatively, an object can be created using the ``with`` keyword. In
        this case, the object is deleted automatically:

        .. code-block:: python

            >>> from pysat.card import ITotalizer
            >>> with ITotalizer(lits=[1, 2, 3], ubound=1) as t:
            ...     print(t.cnf.clauses)
            [[-2, 4], [-1, 4], [-1, -2, 5], [-4, 6], [-5, 7], [-3, 6], [-3, -4, 7]]
            ...     print(t.rhs)
            [6, 7]
    """

    def __init__(self, lits=[], ubound=1, top_id=None):
        """
            Constructor.
        """

        # internal totalizer object
        self.tobj = None

        # its characteristics
        self.lits = []
        self.ubound = 0
        self.top_id = 0

        # encoding result
        self.cnf = CNF()  # CNF formula encoding the totalizer object
        self.rhs = []     # upper bounds on the number of literals (rhs)

        # number of new clauses
        self.nof_new = 0

        # this newly created totalizer object is not yet merged in any other
        self._merged = False

        if lits:
            self.new(lits=lits, ubound=ubound, top_id=top_id)

    def new(self, lits=[], ubound=1, top_id=None):
        """
            The actual constructor of :class:`ITotalizer`. Invoked from
            ``self.__init__()``. Creates an object of :class:`ITotalizer` given
            a list of literals in the sum, the largest potential bound to
            consider, as well as the top variable identifier used so far. See
            the description of :class:`ITotalizer` for details.
        """

        self.lits = list(lits)
        self.ubound = ubound
        self.top_id = max(map(lambda x: abs(x), self.lits + [top_id if top_id != None else 0]))

        if MainThread.check() == True:
            # saving default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_DFL)

            # creating the object
            self.tobj, clauses, self.rhs, self.top_id = pycard.itot_new(self.lits,
                    self.ubound, self.top_id, 1)

            # recovering default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, def_sigint_handler)
        else:
            self.tobj, clauses, self.rhs, self.top_id = pycard.itot_new(self.lits,
                    self.ubound, self.top_id, 0)

        # saving the result
        self.cnf.clauses = clauses
        self.cnf.nv = self.top_id

        # for convenience, keeping the number of clauses
        self.nof_new = len(clauses)

    def delete(self):
        """
            Destroys a previously constructed :class:`ITotalizer` object.
            Internal variables ``self.cnf`` and ``self.rhs`` get cleaned.
        """

        if self.tobj:
            if not self._merged:
                pycard.itot_del(self.tobj)

                # otherwise, this totalizer object is merged into a larger one
                # therefore, this memory should be freed in its destructor

            self.tobj = None

        self.lits = []
        self.ubound = 0
        self.top_id = 0

        self.cnf = CNF()
        self.rhs = []

        self.nof_new = 0

    def __enter__(self):
        """
            'with' constructor.
        """

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
            'with' destructor.
        """

        self.delete()

    def __del__(self):
        """
            Destructor.
        """

        self.delete()

    def increase(self, ubound=1, top_id=None):
        """
            Increases a potential upper bound that can be imposed on the
            literals in the sum of an existing :class:`ITotalizer` object to a
            new value.

            :param ubound: a new upper bound.
            :param top_id: a new top variable identifier.

            :type ubound: int
            :type top_id: integer or None

            The top identifier ``top_id`` applied only if it is greater than
            the one used in ``self``.

            This method creates additional clauses encoding the existing
            totalizer tree up to the new upper bound given and appends them to
            the list of clauses of :class:`.CNF` ``self.cnf``. The number of
            newly created clauses is stored in variable ``self.nof_new``.

            Also, a list of bounds ``self.rhs`` gets increased and its length
            becomes ``ubound+1``.

            The method can be used in the following way:

            .. code-block:: python

                >>> from pysat.card import ITotalizer
                >>> t = ITotalizer(lits=[1, 2, 3], ubound=1)
                >>> print(t.cnf.clauses)
                [[-2, 4], [-1, 4], [-1, -2, 5], [-4, 6], [-5, 7], [-3, 6], [-3, -4, 7]]
                >>> print(t.rhs)
                [6, 7]
                >>>
                >>> t.increase(ubound=2)
                >>> print(t.cnf.clauses)
                [[-2, 4], [-1, 4], [-1, -2, 5], [-4, 6], [-5, 7], [-3, 6], [-3, -4, 7], [-3, -5, 8]]
                >>> print(t.cnf.clauses[-t.nof_new:])
                [[-3, -5, 8]]
                >>> print(t.rhs)
                [6, 7, 8]
                >>> t.delete()
        """

        self.top_id = max(self.top_id, top_id if top_id != None else 0)

        # do nothing if the bound is set incorrectly
        if ubound <= self.ubound or self.ubound >= len(self.lits):
            self.nof_new = 0
            return
        else:
            self.ubound = ubound

        if MainThread.check() == True:
            # saving default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_DFL)

            # updating the object and adding more variables and clauses
            clauses, self.rhs, self.top_id = pycard.itot_inc(self.tobj,
                    self.ubound, self.top_id, 1)

            # recovering default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, def_sigint_handler)
        else:
            clauses, self.rhs, self.top_id = pycard.itot_inc(self.tobj,
                    self.ubound, self.top_id, 0)

        # saving the result
        self.cnf.clauses.extend(clauses)
        self.cnf.nv = self.top_id

        # keeping the number of newly added clauses
        self.nof_new = len(clauses)

    def extend(self, lits=[], ubound=None, top_id=None):
        """
            Extends the list of literals in the sum and (if needed) increases a
            potential upper bound that can be imposed on the complete list of
            literals in the sum of an existing :class:`ITotalizer` object to a
            new value.

            :param lits: additional literals to be included in the sum.
            :param ubound: a new upper bound.
            :param top_id: a new top variable identifier.

            :type lits: iterable(int)
            :type ubound: int
            :type top_id: integer or None

            The top identifier ``top_id`` applied only if it is greater than
            the one used in ``self``.

            This method creates additional clauses encoding the existing
            totalizer tree augmented with new literals in the sum and updating
            the upper bound. As a result, it appends the new clauses to the
            list of clauses of :class:`.CNF` ``self.cnf``. The number of newly
            created clauses is stored in variable ``self.nof_new``.

            Also, if the upper bound is updated, a list of bounds ``self.rhs``
            gets increased and its length becomes ``ubound+1``. Otherwise, it
            is updated with new values.

            The method can be used in the following way:

            .. code-block:: python

                >>> from pysat.card import ITotalizer
                >>> t = ITotalizer(lits=[1, 2], ubound=1)
                >>> print(t.cnf.clauses)
                [[-2, 3], [-1, 3], [-1, -2, 4]]
                >>> print(t.rhs)
                [3, 4]
                >>>
                >>> t.extend(lits=[5], ubound=2)
                >>> print(t.cnf.clauses)
                [[-2, 3], [-1, 3], [-1, -2, 4], [-5, 6], [-3, 6], [-4, 7], [-3, -5, 7], [-4, -5, 8]]
                >>> print(t.cnf.clauses[-t.nof_new:])
                [[-5, 6], [-3, 6], [-4, 7], [-3, -5, 7], [-4, -5, 8]]
                >>> print(t.rhs)
                [6, 7, 8]
                >>> t.delete()
        """

        # preparing a new list of distinct input literals
        lits = list(set(lits).difference(set(self.lits)))

        if not lits:
            # nothing to merge with -> just increase the bound
            if ubound:
                self.increase(ubound=ubound, top_id=top_id)

            return

        self.top_id = max(map(lambda x: abs(x), self.lits + [self.top_id, top_id if top_id != None else 0]))
        self.ubound = max(self.ubound, ubound if ubound != None else 0)

        if MainThread.check() == True:
            # saving default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_DFL)

            # updating the object and adding more variables and clauses
            self.tobj, clauses, self.rhs, self.top_id = pycard.itot_ext(self.tobj,
                    lits, self.ubound, self.top_id, 1)

            # recovering default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, def_sigint_handler)
        else:
            self.tobj, clauses, self.rhs, self.top_id = pycard.itot_ext(self.tobj,
                    lits, self.ubound, self.top_id, 0)

        # saving the result
        self.cnf.clauses.extend(clauses)
        self.cnf.nv = self.top_id
        self.lits.extend(lits)

        # for convenience, keeping the number of new clauses
        self.nof_new = len(clauses)

    def merge_with(self, another, ubound=None, top_id=None):
        """
            This method merges a tree of the current :class:`ITotalizer`
            object, with a tree of another object and (if needed) increases a
            potential upper bound that can be imposed on the complete list of
            literals in the sum of an existing :class:`ITotalizer` object to a
            new value.

            :param another: another totalizer to merge with.
            :param ubound: a new upper bound.
            :param top_id: a new top variable identifier.

            :type another: :class:`ITotalizer`
            :type ubound: int
            :type top_id: integer or None

            The top identifier ``top_id`` applied only if it is greater than
            the one used in ``self``.

            This method creates additional clauses encoding the existing
            totalizer tree merged with another totalizer tree into *one* sum
            and updating the upper bound. As a result, it appends the new
            clauses to the list of clauses of :class:`.CNF` ``self.cnf``. The
            number of newly created clauses is stored in variable
            ``self.nof_new``.

            Also, if the upper bound is updated, a list of bounds ``self.rhs``
            gets increased and its length becomes ``ubound+1``. Otherwise, it
            is updated with new values.

            The method can be used in the following way:

            .. code-block:: python

                >>> from pysat.card import ITotalizer
                >>> with ITotalizer(lits=[1, 2], ubound=1) as t1:
                ...     print(t1.cnf.clauses)
                [[-2, 3], [-1, 3], [-1, -2, 4]]
                ...     print(t1.rhs)
                [3, 4]
                ...
                ...     t2 = ITotalizer(lits=[5, 6], ubound=1)
                ...     print(t1.cnf.clauses)
                [[-6, 7], [-5, 7], [-5, -6, 8]]
                ...     print(t1.rhs)
                [7, 8]
                ...
                ...     t1.merge_with(t2)
                ...     print(t1.cnf.clauses)
                [[-2, 3], [-1, 3], [-1, -2, 4], [-6, 7], [-5, 7], [-5, -6, 8], [-7, 9], [-8, 10], [-3, 9], [-4, 10], [-3, -7, 10]]
                ...     print(t1.cnf.clauses[-t1.nof_new:])
                [[-6, 7], [-5, 7], [-5, -6, 8], [-7, 9], [-8, 10], [-3, 9], [-4, 10], [-3, -7, 10]]
                ...     print(t1.rhs)
                [9, 10]
                ...
                ...     t2.delete()
        """

        self.top_id = max(self.top_id, top_id if top_id != None else 0, another.top_id)
        self.ubound = max(self.ubound, ubound if ubound != None else 0, another.ubound)

        # extending the list of input literals
        self.lits.extend(another.lits)

        if MainThread.check() == True:
            # saving default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_DFL)

            # updating the object and adding more variables and clauses
            self.tobj, clauses, self.rhs, self.top_id = pycard.itot_mrg(self.tobj,
                    another.tobj, self.ubound, self.top_id, 1)

            # recovering default SIGINT handler
            def_sigint_handler = signal.signal(signal.SIGINT, def_sigint_handler)
        else:
            self.tobj, clauses, self.rhs, self.top_id = pycard.itot_mrg(self.tobj,
                    another.tobj, self.ubound, self.top_id, 0)

        # saving the result
        self.cnf.clauses.extend(another.cnf.clauses)
        self.cnf.clauses.extend(clauses)
        self.cnf.nv = self.top_id

        # for convenience, keeping the number of new clauses
        self.nof_new = len(another.cnf.clauses) + len(clauses)

        # memory deallocation should not be done for the merged tree
        another._merged = True



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## hornify.py
##
##  Created on: Mar 21, 2017
##      Author: Alexey S. Ignatiev
##      E-mail: aignatiev@ciencias.ulisboa.pt
##

#
#==============================================================================
from __future__ import print_function
import getopt
import os
from pysat.formula import CNF, WCNF
from six.moves import range
import sys


#
#==============================================================================
class Hornify(WCNF, object):
    """
        Dual-rail encoder creating Horn MaxSAT formulas given a CNF.
    """

    def __init__(self, cnf, nop=False):
        """
            Constructor.
        """

        # initializing WCNF's internal parameters
        super(Hornify, self).__init__()

        self.nv = 2 * cnf.nv
        self.soft = [[l] for l in range(1, self.nv + 1)]
        self.wght = [1 for cl in self.soft]
        self.topw = self.nv + 1

        if not nop:
            for v in range(1, cnf.nv + 1):
                self.hard.append([-v, -v - cnf.nv])

        for cl in cnf.clauses:
            self.hard.append(map(lambda l: l if l < 0 else -l - cnf.nv, cl))


#
#==============================================================================
def parse_options():
    """
        Parses command-line options:
    """

    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   'hp',
                                   ['help',
                                    'nop'])
    except getopt.GetoptError as err:
        sys.stderr.write(str(err).capitalize())
        usage()
        sys.exit(1)

    nop = False

    for opt, arg in opts:
        if opt in ('-h', '--help'):
            usage()
            sys.exit(0)
        elif opt in ('-p', '--nop'):
            nop = True
        else:
            assert False, 'Unhandled option: {0} {1}'.format(opt, arg)

    return nop, args


#
#==============================================================================
def usage():
    """
        Prints usage message.
    """

    print('Usage:', os.path.basename(sys.argv[0]), '[options] dimacs-file')
    print('Options:')
    print('        -h, --help')
    print('        -p, --nop     Discard the P clauses')


#
#==============================================================================
if __name__ == '__main__':
    nop, files = parse_options()
    cnf = CNF()  # input CNF formula

    if files:
        cnf.from_file(files[0])
    else:
        cnf.from_fp(sys.stdin)

    wcnf = Hornify(cnf, nop)
    wcnf.to_fp(sys.stdout)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## _utils.py
##
##  Created on: Nov 27, 2019
##      Author: Alexey S. Ignatiev
##      E-mail: aignatiev@ciencias.ulisboa.pt
##

"""
    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        MainThread

    ==================
    Module description
    ==================

    This simple module is supposed to implement auxiliary classes and routines
    that are used by the other (main) modules of PySAT.

    ==============
    Module details
    ==============
"""

#
#==============================================================================
import threading


#
#==============================================================================
class MainThread(object):
    """
        A dummy class for checking whether the current thread is the main one.
        This is currently necessary for proper signal handling when making
        oracle calls and creating cardinality encodings.
    """

    @staticmethod
    def check():
        """
            The actual checker.
        """

        try:  # for Python > 3.4
            res = threading.current_thread() is threading.main_thread()
        except AttributeError:
            res = isinstance(threading.current_thread(), threading._MainThread)

        return res



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## usage.py
##
##  Created on: Nov 27, 2016
##      Author: Alexey S. Ignatiev
##      E-mail: aignatiev@ciencias.ulisboa.pt
##

#
#==============================================================================
from __future__ import print_function
from pysat.solvers import Solver  # standard way to import the library
from pysat.solvers import Minisat22, Glucose3  # more direct way

#
#==============================================================================
if __name__ == '__main__':
    print('simple infrastructure (add_clause, solve):')
    s1 = Solver(name='g3')
    s1.add_clause([-1, 2])
    s1.add_clause([-2, 3])
    s1.add_clause([-3, 4])

    if s1.solve() == True:
        print(s1.get_model())

    print('blocking models:')
    m = s1.get_model()
    s1.add_clause([-l for l in m])  # actual blocking

    if s1.solve() == True:
        print(s1.get_model())  # should compute another model
    s1.delete()  # this is needed if used not within the 'with' context

    print('model enumeration + bootstrapping the solver:')
    s2 = Minisat22(bootstrap_with=[[-1, 2], [-2, 3], [-3, 4]])  # also working with MiniSat22 directly
    for m in s2.enum_models():  # implemented as a generator
        print(m)
    s2.delete()

    print('\'with\' constructor + assumptions:')
    with Solver(name='mgh', bootstrap_with=[[-1, 2], [-2, 3], [-3, 4]]) as s3:
        for m in s3.enum_models(assumptions=[5]):
            print(m)

        # no s3.delete() is needed -- it is called automatically if used in the 'with' environment

    print('appending formula and extracting an unsat core:')
    with Glucose3(bootstrap_with=[[-1, 2], [-2, 3]]) as s4:
        s4.append_formula([[-3, 4], [-4, 5]])

        if s4.solve(assumptions=[1, -5]) == False:
            print(s4.get_core())

        # no s4.delete() is needed -- it is called automatically if used in the 'with' environment

    print('extracting a proof:')  # supported by glucose and lingeling only
    with Glucose3(bootstrap_with=[[-1, 2], [1, -2], [-1, -2], [1, 2]], with_proof=True) as s5:

        if s5.solve() == False:
            print(s5.get_proof())



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## formula.py
##
##  Created on: Dec 7, 2016
##      Author: Alexey S. Ignatiev
##      E-mail: aignatiev@ciencias.ulisboa.pt
##

"""
    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        IDPool
        CNF
        CNFPlus
        WCNF
        WCNFPlus

    ==================
    Module description
    ==================

    This module is designed to facilitate fast and easy PySAT-development by
    providing a simple way to manipulate formulas in PySAT. Although only
    clausal formulas are supported at this point, future releases of PySAT are
    expected to implement data structures and methods to manipulate arbitrary
    Boolean formulas. The module implements the :class:`CNF` class, which
    represents a formula in `conjunctive normal form (CNF)
    <https://en.wikipedia.org/wiki/Conjunctive_normal_form>`__.

    Recall that a CNF formula is conventionally seen as a set of clauses, each
    being a set of literals. A literal is a Boolean variable or its negation.
    In PySAT, a Boolean variable and a literal should be specified as an
    integer. For instance, a Boolean variable :math:`x_{25}` is represented as
    integer ``25``. A literal :math:`\\neg{x_{10}}` should be specified as
    ``-10``. Moreover, a clause :math:`(\\neg{x_2}\\vee x_{19}\\vee x_{46})`
    should be specified as ``[-2, 19, 46]`` in PySAT. *Unit size clauses* are
    to be specified as unit size lists as well, e.g. a clause :math:`(x_3)` is
    a list ``[3]``.

    CNF formulas can be created as an object of class :class:`CNF`. For
    instance, the following piece of code creates a CNF formula
    :math:`(\\neg{x_1}\\vee x_2)\\wedge(\\neg{x_2}\\vee x_3)`.

    .. code-block:: python

        >>> from pysat.formula import CNF
        >>> cnf = CNF()
        >>> cnf.append([-1, 2])
        >>> cnf.append([-2, 3])

    The clauses of a formula can be accessed through the ``clauses`` variable
    of class :class:`CNF`, which is a list of lists of integers:

    .. code-block:: python

        >>> print(cnf.clauses)
        [[-1, 2], [-2 ,3]]

    The number of variables in a CNF formula, i.e. the *largest variable
    identifier*, can be obtained using the ``nv`` variable, e.g.

    .. code-block:: python

        >>> print(cnf.nv)
        3

    Class :class:`CNF` has a few methods to read and write a CNF formula into a
    file or a string. The formula is read/written in the standard `DIMACS CNF
    <https://en.wikipedia.org/wiki/Boolean_satisfiability_problem#SAT_problem_format>`__
    format. A clause in the DIMACS format is a string containing
    space-separated integer literals  followed by ``0``. For instance, a clause
    :math:`(\\neg{x_2}\\vee x_{19}\\vee x_{46})` is written as ``-2 19 46 0``
    in DIMACS. The clauses in DIMACS should be preceded by a *preamble*, which
    is a line ``p cnf nof_variables nof_clauses``, where ``nof_variables`` and
    ``nof_clauses`` are integers. A preamble line for formula
    :math:`(\\neg{x_1}\\vee x_2)\\wedge(\\neg{x_2}\\vee x_3)` would be ``p cnf
    3 2``. The complete DIMACS file describing the formula looks this:

    ::

        p cnf 3 2
        -1 2 0
        -2 3 0

    Reading and writing formulas in DIMACS can be done with PySAT in the
    following way:

    .. code-block:: python

        >>> from pysat.formula import CNF
        >>> f1 = CNF(from_file='some-file-name.cnf')  # reading from file
        >>> f1.to_file('another-file-name.cnf')  # writing to a file
        >>>
        >>> with open('some-file-name.cnf', 'r+') as fp:
        ...     f2 = CNF(from_fp=fp)  # reading from a file pointer
        ...
        ...     fp.seek(0)
        ...     f2.to_fp(fp)  # writing to a file pointer
        >>>
        >>> f3 = CNF(from_string='p cnf 3 3\\n-1 2 0\\n-2 3 0\\n-3 0\\n')
        >>> print(f3.clauses)
        [[-1, 2], [-2, 3], [-3]]
        >>> print(f3.nv)
        3

    Besides plain CNF formulas, the :mod:`pysat.formula` module implements an
    additional class for dealing with *partial* and *weighted partial* CNF
    formulas, i.e. WCNF formulas. A WCNF formula is a conjunction of two sets
    of clauses: *hard* clauses and *soft* clauses, i.e.
    :math:`\mathcal{F}=\mathcal{H}\wedge\mathcal{S}`. Soft clauses of a WCNF
    are labeled with integer *weights*, i.e. a soft clause of
    :math:`\mathcal{S}` is a pair :math:`(c_i, w_i)`. In partial (unweighted)
    formulas, all soft clauses have weight 1.

    WCNF can be of help when solving optimization problems using the SAT
    technology. A typical example of where a WCNF formula can be used is
    `maximum satisfiability (MaxSAT)
    <https://en.wikipedia.org/wiki/Maximum_satisfiability_problem>`__, which
    given a WCNF formula :math:`\mathcal{F}=\mathcal{H}\wedge\mathcal{S}`
    targets satisfying all its hard clauses :math:`\mathcal{H}` and maximizing
    the sum of weights of satisfied soft clauses, i.e. maximizing the value of
    :math:`\sum_{c_i\in\mathcal{S}}{w_i\\cdot c_i}`.

    An object of class :class:`WCNF` has two variables to access the hard and
    soft clauses of the corresponding formula: ``hard`` and ``soft``. The
    weights of soft clauses are stored in variable ``wght``.

    .. code-block:: python

        >>> from pysat.formula import WCNF
        >>>
        >>> wcnf = WCNF()
        >>> wcnf.append([-1, -2])
        >>> wcnf.append([1], weight=1)
        >>> wcnf.append([2], weight=3)  # the formula becomes unsatisfiable
        >>>
        >>> print(wcnf.hard)
        [[-1, -2]]
        >>> print(wcnf.soft)
        [[1], [2]]
        >>> print(wcnf.wght)
        [1, 3]

    A properly constructed WCNF formula must have a *top weight*, which should
    be equal to :math:`1+\sum_{c_i\in\mathcal{S}}{w_i}`. Top weight of a
    formula can be accessed through variable ``topw``.

    .. code-block:: python

        >>> wcnf.topw = sum(wcnf.wght) + 1  # (1 + 3) + 1
        >>> print(wcnf.topw)
        5

    .. note::

        Although it is not aligned with the WCNF format description, starting
        with the 0.1.5.dev8 release, PySAT is able to deal with WCNF formulas
        having not only integer clause weights but also weights represented as
        *floating point numbers*. Moreover, *negative weights* are also
        supported.

    Additionally to classes :class:`CNF` and :class:`WCNF`, the module provides
    the extended classes :class:`CNFPlus` and :class:`WCNFPlus`. The only
    difference between ``?CNF`` and ``?CNFPlus`` is the support for *native*
    cardinality constraints provided by the `MiniCard solver
    <https://github.com/liffiton/minicard>`__ (see :mod:`pysat.card` for
    details). The corresponding variable in objects of ``CNFPlus``
    (``WCNFPlus``, resp.) responsible for storing the AtMostK constraints is
    ``atmosts`` (``atms``, resp.). **Note** that at this point, AtMostK
    constraints in ``WCNF`` can be *hard* only.

    Besides the implementations of CNF and WCNF formulas in PySAT, the
    :mod:`pysat.formula` module also provides a way to manage variable
    identifiers. This can be done with the use of the :class:`IDPool` manager.
    With the use of the :class:`CNF` and :class:`WCNF` classes as well as with
    the :class:`IDPool` variable manager, it is pretty easy to develop
    practical problem encoders into SAT or MaxSAT/MinSAT. As an example, a PHP
    formula encoder is shown below (the implementation can also be found in
    :class:`.examples.genhard.PHP`).

    .. code-block:: python

        from pysat.formula import CNF
        cnf = CNF()  # we will store the formula here

        # nof_holes is given

        # initializing the pool of variable ids
        vpool = IDPool(start_from=1)
        pigeon = lambda i, j: vpool.id('pigeon{0}@{1}'.format(i, j))

        # placing all pigeons into holes
        for i in range(1, nof_holes + 2):
            cnf.append([pigeon(i, j) for j in range(1, nof_holes + 1)])

        # there cannot be more than 1 pigeon in a hole
        pigeons = range(1, nof_holes + 2)
        for j in range(1, nof_holes + 1):
            for comb in itertools.combinations(pigeons, 2):
                cnf.append([-pigeon(i, j) for i in comb])

    ==============
    Module details
    ==============
"""

#
#==============================================================================
from __future__ import print_function
import collections
import copy
import decimal
import itertools
import os
from pysat._fileio import FileObject
import sys

# checking whether or not py-aiger-cnf is available and working as expected
aiger_present = True
try:
    import aiger_cnf
except ImportError:
    aiger_present = False

try:  # for Python2
    from cStringIO import StringIO
except ImportError:  # for Python3
    from io import StringIO


#
#==============================================================================
class IDPool(object):
    """
        A simple manager of variable IDs. It can be used as a pool of integers
        assigning an ID to any object. Identifiers are to start from ``1`` by
        default. The list of occupied intervals is empty be default. If
        necessary the top variable ID can be accessed directly using the
        ``top`` variable.

        :param start_from: the smallest ID to assign.
        :param occupied: a list of occupied intervals.

        :type start_from: int
        :type occupied: list(list(int))
    """

    def __init__(self, start_from=1, occupied=[]):
        """
            Constructor.
        """

        self.restart(start_from=start_from, occupied=occupied)

    def restart(self, start_from=1, occupied=[]):
        """
            Restart the manager from scratch. The arguments replicate those of
            the constructor of :class:`IDPool`.
        """

        # initial ID
        self.top = start_from - 1

        # occupied IDs
        self._occupied = sorted(occupied, key=lambda x: x[0])

        # main dictionary storing the mapping from objects to variable IDs
        self.obj2id = collections.defaultdict(lambda: self._next())

        # mapping back from variable IDs to objects
        # (if for whatever reason necessary)
        self.id2obj = {}

    def id(self, obj):
        """
            The method is to be used to assign an integer variable ID for a
            given new object. If the object already has an ID, no new ID is
            created and the old one is returned instead.

            An object can be anything. In some cases it is convenient to use
            string variable names.

            :param obj: an object to assign an ID to.

            :rtype: int.

            Example:

            .. code-block:: python

                >>> from pysat.formula import IDPool
                >>> vpool = IDPool(occupied=[[12, 18], [3, 10]])
                >>>
                >>> # creating 5 unique variables for the following strings
                >>> for i in range(5):
                ...    print(vpool.id('v{0}'.format(i + 1)))
                1
                2
                11
                19
                20

            In some cases, it makes sense to create an external function for
            accessing IDPool, e.g.:

            .. code-block:: python

                >>> # continuing the previous example
                >>> var = lambda i: vpool.id('var{0}'.format(i))
                >>> var(5)
                20
                >>> var('hello_world!')
                21
        """

        vid = self.obj2id[obj]

        if vid not in self.id2obj:
            self.id2obj[vid] = obj

        return vid

    def obj(self, vid):
        """
            The method can be used to map back a given variable identifier to
            the original object labeled by the identifier.

            :param vid: variable identifier.
            :type vid: int

            :return: an object corresponding to the given identifier.

            Example:

            .. code-block:: python

                >>> vpool.obj(21)
                'hello_world!'
        """

        if vid in self.id2obj:
            return self.id2obj[vid]

        return None

    def occupy(self, start, stop):
        """
            Mark a given interval as occupied so that the manager could skip
            the values from ``start`` to ``stop`` (**inclusive**).

            :param start: beginning of the interval.
            :param stop: end of the interval.

            :type start: int
            :type stop: int
        """

        self._occupied.append([start, stop])
        self._occupied.sort(key=lambda x: x[0])

    def _next(self):
        """
            Get next variable ID. Skip occupied intervals if any.
        """

        self.top += 1

        while self._occupied and self.top >= self._occupied[0][0]:
            if self.top <= self._occupied[0][1]:
                self.top = self._occupied[0][1] + 1

            self._occupied.pop(0)

        return self.top


#
#==============================================================================
class CNF(object):
    """
        Class for manipulating CNF formulas. It can be used for creating
        formulas, reading them from a file, or writing them to a file. The
        ``comment_lead`` parameter can be helpful when one needs to parse
        specific comment lines starting not with character ``c`` but with
        another character or a string.

        :param from_file: a DIMACS CNF filename to read from
        :param from_fp: a file pointer to read from
        :param from_string: a string storing a CNF formula
        :param from_clauses: a list of clauses to bootstrap the formula with
        :param from_aiger: an AIGER circuit to bootstrap the formula with
        :param comment_lead: a list of characters leading comment lines

        :type from_file: str
        :type from_fp: file_pointer
        :type from_string: str
        :type from_clauses: list(list(int))
        :type from_aiger: :class:`aiger.AIG` (see `py-aiger package <https://github.com/mvcisback/py-aiger>`__)
        :type comment_lead: list(str)
    """

    def __init__(self, from_file=None, from_fp=None, from_string=None,
            from_clauses=[], from_aiger=None, comment_lead=['c']):
        """
            Constructor.
        """

        self.nv = 0
        self.clauses = []
        self.comments = []

        if from_file:
            self.from_file(from_file, comment_lead, compressed_with='use_ext')
        elif from_fp:
            self.from_fp(from_fp, comment_lead)
        elif from_string:
            self.from_string(from_string, comment_lead)
        elif from_clauses:
            self.from_clauses(from_clauses)
        elif from_aiger:
            self.from_aiger(from_aiger)

    def from_file(self, fname, comment_lead=['c'], compressed_with='use_ext'):
        """
            Read a CNF formula from a file in the DIMACS format. A file name is
            expected as an argument. A default argument is ``comment_lead`` for
            parsing comment lines. A given file can be compressed by either
            gzip, bzip2, or lzma.

            :param fname: name of a file to parse.
            :param comment_lead: a list of characters leading comment lines
            :param compressed_with: file compression algorithm

            :type fname: str
            :type comment_lead: list(str)
            :type compressed_with: str

            Note that the ``compressed_with`` parameter can be ``None`` (i.e.
            the file is uncompressed), ``'gzip'``, ``'bzip2'``, ``'lzma'``, or
            ``'use_ext'``. The latter value indicates that compression type
            should be automatically determined based on the file extension.
            Using ``'lzma'`` in Python 2 requires the ``backports.lzma``
            package to be additionally installed.

            Usage example:

            .. code-block:: python

                >>> from pysat.formula import CNF
                >>> cnf1 = CNF()
                >>> cnf1.from_file('some-file.cnf.gz', compressed_with='gzip')
                >>>
                >>> cnf2 = CNF(from_file='another-file.cnf')
        """

        with FileObject(fname, mode='r', compression=compressed_with) as fobj:
            self.from_fp(fobj.fp, comment_lead)

    def from_fp(self, file_pointer, comment_lead=['c']):
        """
            Read a CNF formula from a file pointer. A file pointer should be
            specified as an argument. The only default argument is
            ``comment_lead``, which can be used for parsing specific comment
            lines.

            :param file_pointer: a file pointer to read the formula from.
            :param comment_lead: a list of characters leading comment lines

            :type file_pointer: file pointer
            :type comment_lead: list(str)

            Usage example:

            .. code-block:: python

                >>> with open('some-file.cnf', 'r') as fp:
                ...     cnf1 = CNF()
                ...     cnf1.from_fp(fp)
                >>>
                >>> with open('another-file.cnf', 'r') as fp:
                ...     cnf2 = CNF(from_fp=fp)
        """

        self.nv = 0
        self.clauses = []
        self.comments = []
        comment_lead = tuple('p') + tuple(comment_lead)

        for line in file_pointer:
            line = line.strip()
            if line:
                if line[0] not in comment_lead:
                    cl = [int(l) for l in line.split()[:-1]]
                    self.nv = max([abs(l) for l in cl] + [self.nv])

                    self.clauses.append(cl)
                elif not line.startswith('p cnf '):
                    self.comments.append(line)

    def from_string(self, string, comment_lead=['c']):
        """
            Read a CNF formula from a string. The string should be specified as
            an argument and should be in the DIMACS CNF format. The only
            default argument is ``comment_lead``, which can be used for parsing
            specific comment lines.

            :param string: a string containing the formula in DIMACS.
            :param comment_lead: a list of characters leading comment lines

            :type string: str
            :type comment_lead: list(str)

            Example:

            .. code-block:: python

                >>> from pysat.formula import CNF
                >>> cnf1 = CNF()
                >>> cnf1.from_string(='p cnf 2 2\\n-1 2 0\\n1 -2 0')
                >>> print(cnf1.clauses)
                [[-1, 2], [1, -2]]
                >>>
                >>> cnf2 = CNF(from_string='p cnf 3 3\\n-1 2 0\\n-2 3 0\\n-3 0\\n')
                >>> print(cnf2.clauses)
                [[-1, 2], [-2, 3], [-3]]
                >>> print(cnf2.nv)
                3
        """

        self.from_fp(StringIO(string), comment_lead)

    def from_clauses(self, clauses):
        """
            This methods copies a list of clauses into a CNF object.

            :param clauses: a list of clauses
            :type clauses: list(list(int))

            Example:

            .. code-block:: python

                >>> from pysat.formula import CNF
                >>> cnf = CNF(from_clauses=[[-1, 2], [1, -2], [5]])
                >>> print(cnf.clauses)
                [[-1, 2], [1, -2], [5]]
                >>> print(cnf.nv)
                5
        """

        self.clauses = copy.deepcopy(clauses)

        for cl in self.clauses:
            self.nv = max([abs(l) for l in cl] + [self.nv])

    def from_aiger(self, aig, vpool=None):
        """

            Create a CNF formula by Tseitin-encoding an input AIGER circuit.

            Input circuit is expected to be an object of class
            :class:`aiger.AIG`. Alternatively, it can be specified as an
            :class:`aiger.BoolExpr`, or an ``*.aag`` filename, or an AIGER
            string to parse. (Classes :class:`aiger.AIG` and
            :class:`aiger.BoolExpr` are defined in the `py-aiger package
            <https://github.com/mvcisback/py-aiger>`__.)

            :param aig: an input AIGER circuit
            :param vpool: pool of variable identifiers (optional)

            :type aig: :class:`aiger.AIG` (see `py-aiger package <https://github.com/mvcisback/py-aiger>`__)
            :type vpool: :class:`.IDPool`

            Example:

            .. code-block:: python

                >>> import aiger
                >>> x, y, z = aiger.atom('x'), aiger.atom('y'), aiger.atom('z')
                >>> expr = ~(x | y) & z
                >>> print(expr.aig)
                aag 5 3 0 1 2
                2
                4
                8
                10
                6 3 5
                10 6 8
                i0 y
                i1 x
                i2 z
                o0 6c454aea-c9e1-11e9-bbe3-3af9d34370a9
                >>>
                >>> from pysat.formula import CNF
                >>> cnf = CNF(from_aiger=expr.aig)
                >>> print(cnf.nv)
                5
                >>> print(cnf.clauses)
                [[3, 2, 4], [-3, -4], [-2, -4], [-4, -1, 5], [4, -5], [1, -5]]
                >>> print(['{0} <-> {1}'.format(v, cnf.vpool.obj(v)) for v in cnf.inps])
                ['3 <-> y', '2 <-> x', '1 <-> z']
                >>> print(['{0} <-> {1}'.format(v, cnf.vpool.obj(v)) for v in cnf.outs])
                ['5 <-> 6c454aea-c9e1-11e9-bbe3-3af9d34370a9']
        """

        assert aiger_present, 'Package \'py-aiger-cnf\' is unavailable. Check your installation.'

        # creating a pool of variable IDs if necessary
        self.vpool = vpool if vpool else IDPool()

        # Use py-aiger-cnf to insulate from internal py-aiger details.
        aig_cnf = aiger_cnf.aig2cnf(aig, fresh=self.vpool.id, force_true=False)

        self.clauses = [list(cls) for cls in aig_cnf.clauses]
        self.comments = ['c ' + c.strip() for c in aig_cnf.comments]
        self.nv = max(map(abs, itertools.chain(*self.clauses)))

        # saving input and output variables
        self.inps = list(aig_cnf.input2lit.values())
        self.outs = list(aig_cnf.output2lit.values())

    def copy(self):
        """
            This method can be used for creating a copy of a CNF object. It
            creates another object of the :class:`CNF` class and makes use of
            the *deepcopy* functionality to copy the clauses.

            :return: an object of class :class:`CNF`.

            Example:

            .. code-block:: python

                >>> cnf1 = CNF(from_clauses=[[-1, 2], [1]])
                >>> cnf2 = cnf1.copy()
                >>> print(cnf2.clauses)
                [[-1, 2], [1]]
                >>> print(cnf2.nv)
                2
        """

        cnf = CNF()
        cnf.nv = self.nv
        cnf.clauses = copy.deepcopy(self.clauses)
        cnf.comments = copy.deepcopy(self.comments)

        return cnf

    def to_file(self, fname, comments=None, compress_with='use_ext'):
        """
            The method is for saving a CNF formula into a file in the DIMACS
            CNF format. A file name is expected as an argument. Additionally,
            supplementary comment lines can be specified in the ``comments``
            parameter. Also, a file can be compressed using either gzip, bzip2,
            or lzma (xz).

            :param fname: a file name where to store the formula.
            :param comments: additional comments to put in the file.
            :param compress_with: file compression algorithm

            :type fname: str
            :type comments: list(str)
            :type compress_with: str

            Note that the ``compress_with`` parameter can be ``None`` (i.e.
            the file is uncompressed), ``'gzip'``, ``'bzip2'``, ``'lzma'``, or
            ``'use_ext'``. The latter value indicates that compression type
            should be automatically determined based on the file extension.
            Using ``'lzma'`` in Python 2 requires the ``backports.lzma``
            package to be additionally installed.

            Example:

            .. code-block:: python

                >>> from pysat.formula import CNF
                >>> cnf = CNF()
                ...
                >>> # the formula is filled with a bunch of clauses
                >>> cnf.to_file('some-file-name.cnf')  # writing to a file
        """

        with FileObject(fname, mode='w', compression=compress_with) as fobj:
            self.to_fp(fobj.fp, comments)

    def to_fp(self, file_pointer, comments=None):
        """
            The method can be used to save a CNF formula into a file pointer.
            The file pointer is expected as an argument. Additionally,
            supplementary comment lines can be specified in the ``comments``
            parameter.

            :param fname: a file name where to store the formula.
            :param comments: additional comments to put in the file.

            :type fname: str
            :type comments: list(str)

            Example:

            .. code-block:: python

                >>> from pysat.formula import CNF
                >>> cnf = CNF()
                ...
                >>> # the formula is filled with a bunch of clauses
                >>> with open('some-file.cnf', 'w') as fp:
                ...     cnf.to_fp(fp)  # writing to the file pointer
        """

        # saving formula's internal comments
        for c in self.comments:
            print(c, file=file_pointer)

        # saving externally specified comments
        if comments:
            for c in comments:
                print(c, file=file_pointer)

        print('p cnf', self.nv, len(self.clauses), file=file_pointer)

        for cl in self.clauses:
            print(' '.join(str(l) for l in cl), '0', file=file_pointer)

    def append(self, clause):
        """
            Add one more clause to CNF formula. This method additionally
            updates the number of variables, i.e. variable ``self.nv``, used in
            the formula.

            :param clause: a new clause to add.
            :type clause: list(int)

            .. code-block:: python

                >>> from pysat.formula import CNF
                >>> cnf = CNF(from_clauses=[[-1, 2], [3]])
                >>> cnf.append([-3, 4])
                >>> print(cnf.clauses)
                [[-1, 2], [3], [-3, 4]]
        """

        self.nv = max([abs(l) for l in clause] + [self.nv])
        self.clauses.append(clause)

    def extend(self, clauses):
        """
            Add several clauses to CNF formula. The clauses should be given in
            the form of list. For every clause in the list, method
            :meth:`append` is invoked.

            :param clauses: a list of new clauses to add.
            :type clauses: list(list(int))

            Example:

            .. code-block:: python

                >>> from pysat.formula import CNF
                >>> cnf = CNF(from_clauses=[[-1, 2], [3]])
                >>> cnf.extend([[-3, 4], [5, 6]])
                >>> print(cnf.clauses)
                [[-1, 2], [3], [-3, 4], [5, 6]]
        """

        for cl in clauses:
            self.append(cl)

    def __iter__(self):
        """
            Iterator over all clauses of the formula.
        """

        for cl in self.clauses:
            yield cl

    def weighted(self):
        """
            This method creates a weighted copy of the internal formula. As a
            result, an object of class :class:`WCNF` is returned. Every clause
            of the CNF formula is *soft* in the new WCNF formula and its weight
            is equal to ``1``. The set of hard clauses of the formula is empty.

            :return: an object of class :class:`WCNF`.

            Example:

            .. code-block:: python

                >>> from pysat.formula import CNF
                >>> cnf = CNF(from_clauses=[[-1, 2], [3, 4]])
                >>>
                >>> wcnf = cnf.weighted()
                >>> print(wcnf.hard)
                []
                >>> print(wcnf.soft)
                [[-1, 2], [3, 4]]
                >>> print(wcnf.wght)
                [1, 1]
        """

        wcnf = WCNF()

        wcnf.nv = self.nv
        wcnf.hard = []
        wcnf.soft = copy.deepcopy(self.clauses)
        wcnf.wght = [1 for cl in wcnf.soft]
        wcnf.topw = len(wcnf.wght) + 1
        wcnf.comments = self.comments[:]

        return wcnf

    def negate(self, topv=None):
        """
            Given a CNF formula :math:`\mathcal{F}`, this method creates a CNF
            formula :math:`\\neg{\mathcal{F}}`. The negation of the formula is
            encoded to CNF with the use of *auxiliary* Tseitin variables [1]_.
            A new CNF formula is returned keeping all the newly introduced
            variables that can be accessed through the ``auxvars`` variable.

            **Note** that the negation of each clause is encoded with one
            auxiliary variable if it is not unit size. Otherwise, no auxiliary
            variable is introduced.

            :param topv: top variable identifier if any.
            :type topv: int

            :return: an object of class :class:`CNF`.

            .. [1] G. S. Tseitin. *On the complexity of derivations in the
                propositional calculus*.  Studies in Mathematics and
                Mathematical Logic, Part II. pp.  115–125, 1968

            .. code-block:: python

                >>> from pysat.formula import CNF
                >>> pos = CNF(from_clauses=[[-1, 2], [3]])
                >>> neg = pos.negate()
                >>> print(neg.clauses)
                [[1, -4], [-2, -4], [-1, 2, 4], [4, -3]]
                >>> print(neg.auxvars)
                [4, -3]
        """

        negated = CNF()

        negated.nv = topv
        if not negated.nv:
            negated.nv = self.nv

        negated.clauses = []
        negated.auxvars = []

        for cl in self.clauses:
            auxv = -cl[0]
            if len(cl) > 1:
                negated.nv += 1
                auxv = negated.nv

                # direct implication
                for l in cl:
                    negated.clauses.append([-l, -auxv])

                # opposite implication
                negated.clauses.append(cl + [auxv])

            # keeping all Tseitin variables
            negated.auxvars.append(auxv)

        negated.clauses.append(negated.auxvars)
        return negated


#
#==============================================================================
class WCNF(object):
    """
        Class for manipulating partial (weighted) CNF formulas. It can be used
        for creating formulas, reading them from a file, or writing them to a
        file. The ``comment_lead`` parameter can be helpful when one needs to
        parse specific comment lines starting not with character ``c`` but with
        another character or a string.

        :param from_file: a DIMACS CNF filename to read from
        :param from_fp: a file pointer to read from
        :param from_string: a string storing a CNF formula
        :param comment_lead: a list of characters leading comment lines

        :type from_file: str
        :type from_fp: file_pointer
        :type from_string: str
        :type comment_lead: list(str)
    """

    def __init__(self, from_file=None, from_fp=None, from_string=None,
            comment_lead=['c']):
        """
            Constructor.
        """

        self.nv = 0
        self.hard = []
        self.soft = []
        self.wght = []
        self.topw = 1
        self.comments = []

        if from_file:
            self.from_file(from_file, comment_lead, compressed_with='use_ext')
        elif from_fp:
            self.from_fp(from_fp, comment_lead)
        elif from_string:
            self.from_string(from_string, comment_lead)

    def from_file(self, fname, comment_lead=['c'], compressed_with='use_ext'):
        """
            Read a WCNF formula from a file in the DIMACS format. A file name
            is expected as an argument. A default argument is ``comment_lead``
            for parsing comment lines. A given file can be compressed by either
            gzip, bzip2, or lzma.

            :param fname: name of a file to parse.
            :param comment_lead: a list of characters leading comment lines
            :param compressed_with: file compression algorithm

            :type fname: str
            :type comment_lead: list(str)
            :type compressed_with: str

            Note that the ``compressed_with`` parameter can be ``None`` (i.e.
            the file is uncompressed), ``'gzip'``, ``'bzip2'``, ``'lzma'``, or
            ``'use_ext'``. The latter value indicates that compression type
            should be automatically determined based on the file extension.
            Using ``'lzma'`` in Python 2 requires the ``backports.lzma``
            package to be additionally installed.

            Usage example:

            .. code-block:: python

                >>> from pysat.formula import WCNF
                >>> cnf1 = WCNF()
                >>> cnf1.from_file('some-file.wcnf.bz2', compressed_with='bzip2')
                >>>
                >>> cnf2 = WCNF(from_file='another-file.wcnf')
        """

        with FileObject(fname, mode='r', compression=compressed_with) as fobj:
            self.from_fp(fobj.fp, comment_lead)

    def from_fp(self, file_pointer, comment_lead=['c']):
        """
            Read a WCNF formula from a file pointer. A file pointer should be
            specified as an argument. The only default argument is
            ``comment_lead``, which can be used for parsing specific comment
            lines.

            :param file_pointer: a file pointer to read the formula from.
            :param comment_lead: a list of characters leading comment lines

            :type file_pointer: file pointer
            :type comment_lead: list(str)

            Usage example:

            .. code-block:: python

                >>> with open('some-file.cnf', 'r') as fp:
                ...     cnf1 = WCNF()
                ...     cnf1.from_fp(fp)
                >>>
                >>> with open('another-file.cnf', 'r') as fp:
                ...     cnf2 = WCNF(from_fp=fp)
        """

        def parse_wght(string):
            wght = float(string)
            return int(wght) if wght.is_integer() else decimal.Decimal(string)

        self.nv = 0
        self.hard = []
        self.soft = []
        self.wght = []
        self.topw = 1
        self.comments = []
        comment_lead = tuple('p') + tuple(comment_lead)

        # soft clauses with negative weights
        negs = []

        for line in file_pointer:
            line = line.strip()
            if line:
                if line[0] not in comment_lead:
                    items = line.split()[:-1]
                    w, cl = parse_wght(items[0]), [int(l) for l in items[1:]]
                    self.nv = max([abs(l) for l in cl] + [self.nv])

                    if w <= 0:
                        # this clause has a negative weight
                        # it will be processed later
                        negs.append(tuple([cl, -w]))
                    elif w >= self.topw:
                        self.hard.append(cl)
                    else:
                        self.soft.append(cl)
                        self.wght.append(w)
                elif not line.startswith('p wcnf '):
                    self.comments.append(line)
                else:  # expecting the preamble
                    self.topw = parse_wght(line.rsplit(' ', 1)[1])

        # if there is any soft clause with negative weight
        # normalize it, i.e. transform into a set of clauses
        # with a positive weight
        if negs:
            self.normalize_negatives(negs)

    def normalize_negatives(self, negatives):
        """
            Iterate over all soft clauses with negative weights and add their
            negation either as a hard clause or a soft one.

            :param negatives: soft clauses with their negative weights.
            :type negatives: list(list(int))
        """

        for cl, w in negatives:
            selv = cl[0]

            # tseitin-encoding the clause if it is not unit-size
            if len(cl) > 1:
                self.nv += 1
                selv = self.nv

                for l in cl:
                    self.hard.append([selv, -l])
                self.hard.append([-selv] + cl)

            # adding the negation of the clause either as hard or soft
            if w >= self.topw:
                self.hard.append([-selv])
            else:
                self.soft.append([-selv])
                self.wght.append(w)

    def from_string(self, string, comment_lead=['c']):
        """
            Read a WCNF formula from a string. The string should be specified
            as an argument and should be in the DIMACS CNF format. The only
            default argument is ``comment_lead``, which can be used for parsing
            specific comment lines.

            :param string: a string containing the formula in DIMACS.
            :param comment_lead: a list of characters leading comment lines

            :type string: str
            :type comment_lead: list(str)

            Example:

            .. code-block:: python

                >>> from pysat.formula import WCNF
                >>> cnf1 = WCNF()
                >>> cnf1.from_string(='p wcnf 2 2 2\\n 2 -1 2 0\\n1 1 -2 0')
                >>> print(cnf1.hard)
                [[-1, 2]]
                >>> print(cnf1.soft)
                [[1, 2]]
                >>>
                >>> cnf2 = WCNF(from_string='p wcnf 3 3 2\\n2 -1 2 0\\n2 -2 3 0\\n1 -3 0\\n')
                >>> print(cnf2.hard)
                [[-1, 2], [-2, 3]]
                >>> print(cnf2.soft)
                [[-3]]
                >>> print(cnf2.nv)
                3
        """

        self.from_fp(StringIO(string), comment_lead)

    def copy(self):
        """
            This method can be used for creating a copy of a WCNF object. It
            creates another object of the :class:`WCNF` class and makes use of
            the *deepcopy* functionality to copy both hard and soft clauses.

            :return: an object of class :class:`WCNF`.

            Example:

            .. code-block:: python

                >>> cnf1 = WCNF()
                >>> cnf1.append([-1, 2])
                >>> cnf1.append([1], weight=10)
                >>>
                >>> cnf2 = cnf1.copy()
                >>> print(cnf2.hard)
                [[-1, 2]]
                >>> print(cnf2.soft)
                [[1]]
                >>> print(cnf2.wght)
                [10]
                >>> print(cnf2.nv)
                2
        """

        wcnf = WCNF()
        wcnf.nv = self.nv
        wcnf.topw = self.topw
        wcnf.hard = copy.deepcopy(self.hard)
        wcnf.soft = copy.deepcopy(self.soft)
        wcnf.wght = copy.deepcopy(self.wght)
        wcnf.comments = copy.deepcopy(self.comments)

        return wcnf

    def to_file(self, fname, comments=None, compress_with='use_ext'):
        """
            The method is for saving a WCNF formula into a file in the DIMACS
            CNF format. A file name is expected as an argument. Additionally,
            supplementary comment lines can be specified in the ``comments``
            parameter. Also, a file can be compressed using either gzip, bzip2,
            or lzma (xz).

            :param fname: a file name where to store the formula.
            :param comments: additional comments to put in the file.
            :param compress_with: file compression algorithm

            :type fname: str
            :type comments: list(str)
            :type compress_with: str

            Note that the ``compress_with`` parameter can be ``None`` (i.e.
            the file is uncompressed), ``'gzip'``, ``'bzip2'``, ``'lzma'``, or
            ``'use_ext'``. The latter value indicates that compression type
            should be automatically determined based on the file extension.
            Using ``'lzma'`` in Python 2 requires the ``backports.lzma``
            package to be additionally installed.

            Example:

            .. code-block:: python

                >>> from pysat.formula import WCNF
                >>> wcnf = WCNF()
                ...
                >>> # the formula is filled with a bunch of clauses
                >>> wcnf.to_file('some-file-name.wcnf')  # writing to a file
        """

        with FileObject(fname, mode='w', compression=compress_with) as fobj:
            self.to_fp(fobj.fp, comments)

    def to_fp(self, file_pointer, comments=None):
        """
            The method can be used to save a WCNF formula into a file pointer.
            The file pointer is expected as an argument. Additionally,
            supplementary comment lines can be specified in the ``comments``
            parameter.

            :param fname: a file name where to store the formula.
            :param comments: additional comments to put in the file.

            :type fname: str
            :type comments: list(str)

            Example:

            .. code-block:: python

                >>> from pysat.formula import WCNF
                >>> wcnf = WCNF()
                ...
                >>> # the formula is filled with a bunch of clauses
                >>> with open('some-file.wcnf', 'w') as fp:
                ...     wcnf.to_fp(fp)  # writing to the file pointer
        """

        # saving formula's internal comments
        for c in self.comments:
            print(c, file=file_pointer)

        # saving externally specified comments
        if comments:
            for c in comments:
                print(c, file=file_pointer)

        print('p wcnf', self.nv, len(self.hard) + len(self.soft), self.topw, file=file_pointer)

        # soft clauses are dumped first because
        # some tools (e.g. LBX) cannot count them properly
        for i, cl in enumerate(self.soft):
            print(self.wght[i], ' '.join(str(l) for l in cl), '0', file=file_pointer)

        for cl in self.hard:
            print(self.topw, ' '.join(str(l) for l in cl), '0', file=file_pointer)

    def append(self, clause, weight=None):
        """
            Add one more clause to WCNF formula. This method additionally
            updates the number of variables, i.e. variable ``self.nv``, used in
            the formula.

            The clause can be hard or soft depending on the ``weight``
            argument. If no weight is set, the clause is considered to be hard.

            :param clause: a new clause to add.
            :param weight: integer weight of the clause.

            :type clause: list(int)
            :type weight: integer or None

            .. code-block:: python

                >>> from pysat.formula import WCNF
                >>> cnf = WCNF()
                >>> cnf.append([-1, 2])
                >>> cnf.append([1], weight=10)
                >>> cnf.append([-2], weight=20)
                >>> print(cnf.hard)
                [[-1, 2]]
                >>> print(cnf.soft)
                [[1], [-2]]
                >>> print(cnf.wght)
                [10, 20]
        """

        self.nv = max([abs(l) for l in clause] + [self.nv])

        if weight:
            self.soft.append(clause)
            self.wght.append(weight)

            self.topw += weight
        else:
            self.hard.append(clause)

    def extend(self, clauses, weights=None):
        """
            Add several clauses to WCNF formula. The clauses should be given in
            the form of list. For every clause in the list, method
            :meth:`append` is invoked.

            The clauses can be hard or soft depending on the ``weights``
            argument. If no weights are set, the clauses are considered to be
            hard.

            :param clauses: a list of new clauses to add.
            :param weights: a list of integer weights.

            :type clauses: list(list(int))
            :type weights: list(int)

            Example:

            .. code-block:: python

                >>> from pysat.formula import WCNF
                >>> cnf = WCNF()
                >>> cnf.extend([[-3, 4], [5, 6]])
                >>> cnf.extend([[3], [-4], [-5], [-6]], weights=[1, 5, 3, 4])
                >>> print(cnf.hard)
                [[-3, 4], [5, 6]]
                >>> print(cnf.soft)
                [[3], [-4], [-5], [-6]]
                >>> print(cnf.wght)
                [1, 5, 3, 4]
        """

        if weights:
            # clauses are soft
            for i, cl in enumerate(clauses):
                self.append(cl, weight=weights[i])
        else:
            # clauses are hard
            for cl in clauses:
                self.append(cl)

    def unweighted(self):
        """
            This method creates a *plain* (unweighted) copy of the internal
            formula. As a result, an object of class :class:`CNF` is returned.
            Every clause (both hard or soft) of the WCNF formula is copied to
            the ``clauses`` variable of the resulting plain formula, i.e. all
            weights are discarded.

            :return: an object of class :class:`CNF`.

            Example:

            .. code-block:: python

                >>> from pysat.formula import WCNF
                >>> wcnf = WCNF()
                >>> wcnf.extend([[-3, 4], [5, 6]])
                >>> wcnf.extend([[3], [-4], [-5], [-6]], weights=[1, 5, 3, 4])
                >>>
                >>> cnf = wcnf.unweighted()
                >>> print(cnf.clauses)
                [[-3, 4], [5, 6], [3], [-4], [-5], [-6]]
        """

        cnf = CNF()

        cnf.nv = self.nv
        cnf.clauses = copy.deepcopy(self.hard) + copy.deepcopy(self.soft)
        cnf.commends = self.comments[:]

        return cnf


#
#==============================================================================
class CNFPlus(CNF, object):
    """
        CNF formulas augmented with *native* cardinality constraints.

        This class inherits most of the functionality of the :class:`CNF`
        class. The only difference between the two is that :class:`CNFPlus`
        supports *native* cardinality constraints of `MiniCard
        <https://github.com/liffiton/minicard>`__.

        The parser of input DIMACS files of :class:`CNFPlus` assumes the syntax
        of AtMostK and AtLeastK constraints defined in the `description
        <https://github.com/liffiton/minicard>`__ of MiniCard:

        ::

            c Example: Two cardinality constraints followed by a clause
            p cnf+ 7 3
            1 -2 3 5 -7 <= 3
            4 5 6 -7 >= 2
            3 5 7 0

        Each AtLeastK constraint is translated into an AtMostK constraint in
        the standard way: :math:`\sum_{i=1}^{n}{x_i}\geq k \leftrightarrow
        \sum_{i=1}^{n}{\\neg{x_i}}\leq (n-k)`. Internally, AtMostK constraints
        are stored in variable ``atmosts``, each being a pair ``(lits, k)``,
        where ``lits`` is a list of literals in the sum and ``k`` is the upper
        bound.

        Example:

        .. code-block:: python

            >>> from pysat.formula import CNFPlus
            >>> cnf = CNFPlus(from_string='p cnf+ 7 3\\n1 -2 3 5 -7 <= 3\\n4 5 6 -7 >= 2\\n 3 5 7 0\\n')
            >>> print(cnf.clauses)
            [[3, 5, 7]]
            >>> print(cnf.atmosts)
            [[[1, -2, 3, 5, -7], 3], [[-4, -5, -6, 7], 2]]
            >>> print(cnf.nv)
            7

        For details on the functionality, see :class:`CNF`.
    """

    def __init__(self, from_file=None, from_fp=None, from_string=None,
            comment_lead=['c']):
        """
            Constructor.
        """

        # atmost constraints are initially empty
        self.atmosts = []

        # calling the base class constructor
        super(CNFPlus, self).__init__(from_file=from_file, from_fp=from_fp,
                from_string=from_string, comment_lead=comment_lead)

    def from_fp(self, file_pointer, comment_lead=['c']):
        """
            Read a CNF+ formula from a file pointer. A file pointer should be
            specified as an argument. The only default argument is
            ``comment_lead``, which can be used for parsing specific comment
            lines.

            :param file_pointer: a file pointer to read the formula from.
            :param comment_lead: a list of characters leading comment lines

            :type file_pointer: file pointer
            :type comment_lead: list(str)

            Usage example:

            .. code-block:: python

                >>> with open('some-file.cnf+', 'r') as fp:
                ...     cnf1 = CNFPlus()
                ...     cnf1.from_fp(fp)
                >>>
                >>> with open('another-file.cnf+', 'r') as fp:
                ...     cnf2 = CNFPlus(from_fp=fp)
        """

        self.nv = 0
        self.clauses = []
        self.atmosts = []
        self.comments = []
        comment_lead = tuple('p') + tuple(comment_lead)

        for line in file_pointer:
            line = line.strip()
            if line:
                if line[0] not in comment_lead:
                    if int(line.rsplit(' ', 1)[-1]) == 0:  # normal clause
                        cl = [int(l) for l in line.split()[:-1]]
                        self.nv = max([abs(l) for l in cl] + [self.nv])

                        self.clauses.append(cl)
                    else:  # atmost/atleast constraint
                        items = [i for i in line.split()]
                        lits = [int(l) for l in items[:-2]]
                        rhs = int(items[-1])
                        self.nv = max([abs(l) for l in lits] + [self.nv])

                        if items[-2][0] == '>':
                            lits = list(map(lambda l: -l, lits))
                            rhs = len(lits) - rhs

                        self.atmosts.append([lits, rhs])
                elif not line.startswith('p cnf'):  # cnf is allowed here
                    self.comments.append(line)

    def to_fp(self, file_pointer, comments=None):
        """
            The method can be used to save a CNF+ formula into a file pointer.
            The file pointer is expected as an argument. Additionally,
            supplementary comment lines can be specified in the ``comments``
            parameter.

            :param fname: a file name where to store the formula.
            :param comments: additional comments to put in the file.

            :type fname: str
            :type comments: list(str)

            Example:

            .. code-block:: python

                >>> from pysat.formula import CNFPlus
                >>> cnf = CNFPlus()
                ...
                >>> # the formula is filled with a bunch of clauses
                >>> with open('some-file.cnf+', 'w') as fp:
                ...     cnf.to_fp(fp)  # writing to the file pointer
        """

        # saving formula's internal comments
        for c in self.comments:
            print(c, file=file_pointer)

        # saving externally specified comments
        if comments:
            for c in comments:
                print(c, file=file_pointer)

        ftype = 'cnf+' if self.atmosts else 'cnf'
        print('p', ftype, self.nv, len(self.clauses) + len(self.atmosts),
                file=file_pointer)

        for cl in self.clauses:
            print(' '.join(str(l) for l in cl), '0', file=file_pointer)

        for am in self.atmosts:
            print(' '.join(str(l) for l in am[0]), '<=', am[1], file=file_pointer)

    def append(self, clause, is_atmost=False):
        """
            Add a single clause or a single AtMostK constraint to CNF+ formula.
            This method additionally updates the number of variables, i.e.
            variable ``self.nv``, used in the formula.

            If the clause is an AtMostK constraint, this should be set with the
            use of the additional default argument ``is_atmost``, which is set
            to ``False`` by default.

            :param clause: a new clause to add.
            :param is_atmost: if ``True``, the clause is AtMostK.

            :type clause: list(int)
            :type is_atmost: bool

            .. code-block:: python

                >>> from pysat.formula import CNFPlus
                >>> cnf = CNFPlus()
                >>> cnf.append([-3, 4])
                >>> cnf.append([[1, 2, 3], 1], is_atmost=True)
                >>> print(cnf.clauses)
                [[-3, 4]]
                >>> print(cnf.atmosts)
                [[1, 2, 3], 1]
        """

        if not is_atmost:
            self.nv = max([abs(l) for l in clause] + [self.nv])
            self.clauses.append(clause)
        else:
            self.nv = max([abs(l) for l in clause[0]] + [self.nv])
            self.atmosts.append(clause)

    def weighted(self):
        """
            This method creates a weighted copy of the internal formula. As a
            result, an object of class :class:`WCNFPlus` is returned. Every
            clause of the CNFPlus formula is *soft* in the new WCNFPlus
            formula and its weight is equal to ``1``. The set of hard clauses
            of the new formula is empty. The set of cardinality constraints
            remains unchanged.

            :return: an object of class :class:`WCNFPlus`.

            Example:

            .. code-block:: python

                >>> from pysat.formula import CNFPlus
                >>> cnf = CNFPlus()
                >>> cnf.append([-1, 2])
                >>> cnf.append([3, 4])
                >>> cnf.append([[1, 2], 1], is_atmost=True)
                >>>
                >>> wcnf = cnf.weighted()
                >>> print(wcnf.hard)
                []
                >>> print(wcnf.soft)
                [[-1, 2], [3, 4]]
                >>> print(wcnf.wght)
                [1, 1]
                >>> print(wcnf.atms)
                [[[1, 2], 1]]
        """

        wcnf = WCNFPlus()

        wcnf.nv = self.nv
        wcnf.hard = []
        wcnf.soft = copy.deepcopy(self.clauses)
        wcnf.atms = copy.deepcopy(self.atmosts)
        wcnf.wght = [1 for cl in wcnf.soft]
        wcnf.topw = len(wcnf.wght) + 1
        wcnf.comments = self.comments[:]

        return wcnf

    def copy(self):
        """
            This method can be used for creating a copy of a CNFPlus object.
            It creates another object of the :class:`CNFPlus` class, call the
            copy function of CNF class and makes use of the *deepcopy*
            functionality to copy the atmost constraints.

            :return: an object of class :class:`CNFPlus`.

            Example:

            .. code-block:: python

                >>> cnf1 = CNFPlus()
                >>> cnf1.extend([[-1, 2], [1]])
                >>> cnf1.append([[1, 2], 1], is_atmost=True)
                >>> cnf2 = cnf1.copy()
                >>> print(cnf2.clauses)
                [[-1, 2], [1]]
                >>> print(cnf2.nv)
                2
                >>> print(cnf2.atmosts)
                [[[1, 2], 1]]
        """

        cnfplus = super(CNFPlus, self).copy()
        cnfplus.atmosts = copy.deepcopy(self.atmosts)

        return cnfplus


#
#==============================================================================
class WCNFPlus(WCNF, object):
    """
        WCNF formulas augmented with *native* cardinality constraints.

        This class inherits most of the functionality of the :class:`WCNF`
        class. The only difference between the two is that :class:`WCNFPlus`
        supports *native* cardinality constraints of `MiniCard
        <https://github.com/liffiton/minicard>`__.

        The parser of input DIMACS files of :class:`WCNFPlus` assumes the
        syntax of AtMostK and AtLeastK constraints following the one defined
        for :class:`CNFPlus` in the `description
        <https://github.com/liffiton/minicard>`__ of MiniCard:

        ::

            c Example: Two (hard) cardinality constraints followed by a soft clause
            p wcnf+ 7 3 10
            10 1 -2 3 5 -7 <= 3
            10 4 5 6 -7 >= 2
            5 3 5 7 0

        **Note** that every cardinality constraint is assumed to be *hard*,
        i.e. soft cardinality constraints are currently *not supported*.

        Each AtLeastK constraint is translated into an AtMostK constraint in
        the standard way: :math:`\sum_{i=1}^{n}{x_i}\geq k \leftrightarrow
        \sum_{i=1}^{n}{\\neg{x_i}}\leq (n-k)`. Internally, AtMostK constraints
        are stored in variable ``atms``, each being a pair ``(lits, k)``, where
        ``lits`` is a list of literals in the sum and ``k`` is the upper bound.

        Example:

        .. code-block:: python

            >>> from pysat.formula import WCNFPlus
            >>> cnf = WCNFPlus(from_string='p wcnf+ 7 3 10\\n10 1 -2 3 5 -7 <= 3\\n10 4 5 6 -7 >= 2\\n5 3 5 7 0\\n')
            >>> print(cnf.soft)
            [[3, 5, 7]]
            >>> print(cnf.wght)
            [5]
            >>> print(cnf.hard)
            []
            >>> print(cnf.atms)
            [[[1, -2, 3, 5, -7], 3], [[-4, -5, -6, 7], 2]]
            >>> print(cnf.nv)
            7

        For details on the functionality, see :class:`WCNF`.
    """

    def __init__(self, from_file=None, from_fp=None, from_string=None, comment_lead=['c']):
        """
            Constructor.
        """

        # atmost constraints are initially empty
        self.atms = []

        # calling the base class constructor
        super(WCNFPlus, self).__init__(from_file=from_file, from_fp=from_fp,
                from_string=from_string, comment_lead=comment_lead)

    def from_fp(self, file_pointer, comment_lead=['c']):
        """
            Read a WCNF+ formula from a file pointer. A file pointer should be
            specified as an argument. The only default argument is
            ``comment_lead``, which can be used for parsing specific comment
            lines.

            :param file_pointer: a file pointer to read the formula from.
            :param comment_lead: a list of characters leading comment lines

            :type file_pointer: file pointer
            :type comment_lead: list(str)

            Usage example:

            .. code-block:: python

                >>> with open('some-file.wcnf+', 'r') as fp:
                ...     cnf1 = WCNFPlus()
                ...     cnf1.from_fp(fp)
                >>>
                >>> with open('another-file.wcnf+', 'r') as fp:
                ...     cnf2 = WCNFPlus(from_fp=fp)
        """

        def parse_wght(string):
            wght = float(string)
            return int(wght) if wght.is_integer() else decimal.Decimal(string)

        self.nv = 0
        self.hard = []
        self.atms = []
        self.soft = []
        self.wght = []
        self.topw = 1
        self.comments = []
        comment_lead = tuple('p') + tuple(comment_lead)

        for line in file_pointer:
            line = line.strip()
            if line:
                if line[0] not in comment_lead:
                    if int(line.rsplit(' ', 1)[-1]) == 0:  # normal clause
                        items = line.split()[:-1]
                        w, cl = parse_wght(items[0]), [int(l) for l in items[1:]]
                        self.nv = max([abs(l) for l in cl] + [self.nv])

                        if w <= 0:
                            # this clause has a negative weight
                            # it will be processed later
                            negs.append(tuple([cl, -w]))
                        elif w >= self.topw:
                            self.hard.append(cl)
                        else:
                            self.soft.append(cl)
                            self.wght.append(w)
                    else:  # atmost/atleast constraint
                        items = [i for i in line.split()]
                        lits = [int(l) for l in items[1:-2]]
                        rhs = int(items[-1])
                        self.nv = max([abs(l) for l in lits] + [self.nv])

                        if items[-2][0] == '>':
                            lits = list(map(lambda l: -l, lits))
                            rhs = len(lits) - rhs

                        self.atms.append([lits, rhs])
                elif not line.startswith('p wcnf'):  # wcnf is allowed here
                    self.comments.append(line)
                else:  # expecting the preamble
                    self.topw = parse_wght(line.rsplit(' ', 1)[1])

    def to_fp(self, file_pointer, comments=None):
        """
            The method can be used to save a WCNF+ formula into a file pointer.
            The file pointer is expected as an argument. Additionally,
            supplementary comment lines can be specified in the ``comments``
            parameter.

            :param fname: a file name where to store the formula.
            :param comments: additional comments to put in the file.

            :type fname: str
            :type comments: list(str)

            Example:

            .. code-block:: python

                >>> from pysat.formula import WCNFPlus
                >>> cnf = WCNFPlus()
                ...
                >>> # the formula is filled with a bunch of clauses
                >>> with open('some-file.wcnf+', 'w') as fp:
                ...     cnf.to_fp(fp)  # writing to the file pointer
        """

        # saving formula's internal comments
        for c in self.comments:
            print(c, file=file_pointer)

        # saving externally specified comments
        if comments:
            for c in comments:
                print(c, file=file_pointer)

        ftype = 'wcnf+' if self.atms else 'wcnf'
        print('p', ftype, self.nv, len(self.hard) + len(self.soft) + len(self.atms),
                self.topw, file=file_pointer)

        # soft clauses are dumped first because
        # some tools (e.g. LBX) cannot count them properly
        for i, cl in enumerate(self.soft):
            print(self.wght[i], ' '.join(str(l) for l in cl), '0', file=file_pointer)

        for cl in self.hard:
            print(self.topw, ' '.join(str(l) for l in cl), '0', file=file_pointer)

        # atmost constraints are hard
        for am in self.atms:
            print(self.topw, ' '.join(str(l) for l in am[0]), '<=', am[1], file=file_pointer)

    def append(self, clause, weight=None, is_atmost=False):
        """
            Add a single clause or a single AtMostK constraint to WCNF+
            formula. This method additionally updates the number of variables,
            i.e.  variable ``self.nv``, used in the formula.

            If the clause is an AtMostK constraint, this should be set with the
            use of the additional default argument ``is_atmost``, which is set
            to ``False`` by default.

            If ``is_atmost`` is set to ``False``, the clause can be either hard
            or soft depending on the ``weight`` argument. If no weight is
            specified, the clause is considered hard. Otherwise, the clause is
            soft.

            :param clause: a new clause to add.
            :param weight: an integer weight of the clause.
            :param is_atmost: if ``True``, the clause is AtMostK.

            :type clause: list(int)
            :type weight: integer or None
            :type is_atmost: bool

            .. code-block:: python

                >>> from pysat.formula import WCNFPlus
                >>> cnf = WCNFPlus()
                >>> cnf.append([-3, 4])
                >>> cnf.append([[1, 2, 3], 1], is_atmost=True)
                >>> cnf.append([-1, -2], weight=35)
                >>> print(cnf.hard)
                [[-3, 4]]
                >>> print(cnf.atms)
                [[1, 2, 3], 1]
                >>> print(cnf.soft)
                [[-1, -2]]
                >>> print(cnf.wght)
                [35]
        """

        if not is_atmost:
            self.nv = max([abs(l) for l in clause] + [self.nv])

            if weight:
                self.soft.append(clause)
                self.wght.append(weight)

                self.topw += weight
            else:
                self.hard.append(clause)
        else:
            self.nv = max([abs(l) for l in clause[0]] + [self.nv])
            self.atms.append(clause)

    def unweighted(self):
        """
            This method creates a *plain* (unweighted) copy of the internal
            formula. As a result, an object of class :class:`CNFPlus` is
            returned. Every clause (both hard or soft) of the original
            WCNFPlus formula is copied to the ``clauses`` variable of the
            resulting plain formula, i.e. all weights are discarded.

            Note that the cardinality constraints of the original (weighted)
            formula remain unchanged in the new (plain) formula.

            :return: an object of class :class:`CNFPlus`.

            Example:

            .. code-block:: python

                >>> from pysat.formula import WCNF
                >>> wcnf = WCNFPlus()
                >>> wcnf.extend([[-3, 4], [5, 6]])
                >>> wcnf.extend([[3], [-4], [-5], [-6]], weights=[1, 5, 3, 4])
                >>> wcnf.append([[1, 2, 3], 1], is_atmost=True)
                >>>
                >>> cnf = wcnf.unweighted()
                >>> print(cnf.clauses)
                [[-3, 4], [5, 6], [3], [-4], [-5], [-6]]
                >>> print(cnf.atmosts)
                [[[1, 2, 3], 1]]
        """

        cnf = CNFPlus()

        cnf.nv = self.nv
        cnf.clauses = copy.deepcopy(self.hard) + copy.deepcopy(self.soft)
        cnf.atmosts = copy.deepcopy(self.atms)
        cnf.commends = self.comments[:]

        return cnf

    def copy(self):
        """
            This method can be used for creating a copy of a WCNFPlus object.
            It creates another object of the :class:`WCNFPlus` class, call the
            copy function of WCNF class and makes use of the *deepcopy*
            functionality to copy the atmost constraints.

            :return: an object of class :class:`WCNFPlus`.

            Example:

            .. code-block:: python

                >>> cnf1 = WCNFPlus()
                >>> cnf1.append([-1, 2])
                >>> cnf1.append([1], weight=10)
                >>> cnf1.append([[1, 2], 1], is_atmost=True)
                >>> cnf2 = cnf1.copy()
                >>> print(cnf2.hard)
                [[-1, 2]]
                >>> print(cnf2.soft)
                [[1]]
                >>> print(cnf2.wght)
                [10]
                >>> print(cnf2.nv)
                2
                >> print(cnf2.atms)
                [[[1, 2], 1]]

        """

        wcnfplus = super(WCNFPlus, self).copy()
        wcnfplus.atms = copy.deepcopy(self.atms)

        return wcnfplus



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## setup.py
##
##  Created on: Jan 23, 2018
##      Author: Alexey S. Ignatiev
##      E-mail: aignatiev@ciencias.ulisboa.pt
##

#
#==============================================================================
import os
import os.path
import contextlib
import glob

try:
    from setuptools import setup, Extension
    HAVE_SETUPTOOLS = True
except ImportError:
    from distutils.core import setup, Extension
    HAVE_SETUPTOOLS = False

import distutils.command.build
import distutils.command.install

import inspect, os, sys
sys.path.insert(0, os.path.join(os.path.realpath(os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])), 'solvers/'))
import platform
import prepare

from pysat import __version__


#
#==============================================================================
@contextlib.contextmanager
def chdir(new_dir):
    old_dir = os.getcwd()
    try:
        os.chdir(new_dir)
        yield
    finally:
        os.chdir(old_dir)


#
#==============================================================================
ROOT = os.path.abspath(os.path.dirname(__file__))

LONG_DESCRIPTION = """
A Python library providing a simple interface to a number of state-of-art
Boolean satisfiability (SAT) solvers and a few types of cardinality and
pseudo-Boolean encodings. The purpose of PySAT is to enable researchers
working on SAT and its applications and generalizations to easily prototype
with SAT oracles in Python while exploiting incrementally the power of the
original low-level implementations of modern SAT solvers.

With PySAT it should be easy for you to implement a MaxSAT solver, an
MUS/MCS extractor/enumerator, or any tool solving an application problem
with the (potentially multiple) use of a SAT oracle.

Details can be found at `https://pysathq.github.io <https://pysathq.github.io>`__.
"""


# solvers to install
#==============================================================================
to_install = ['cadical', 'glucose30', 'glucose41', 'lingeling', 'maplechrono',
        'maplecm', 'maplesat', 'minicard', 'minisat22', 'minisatgh']


# example scripts to install as standalone executables
#==============================================================================
scripts = ['fm', 'genhard', 'lbx', 'lsu', 'mcsls', 'models', 'musx', 'rc2']


# we need to redefine the build command to
# be able to download and compile solvers
#==============================================================================
class build(distutils.command.build.build):
    """
        Our custom builder class.
    """

    def run(self):
        """
            Download, patch and compile SAT solvers before building.
        """
        # download and compile solvers
        prepare.do(to_install)

        # now, do standard build
        distutils.command.build.build.run(self)


# compilation flags for C extensions
#==============================================================================
compile_flags, cpplib = ['-std=c++11', '-Wall', '-Wno-deprecated'],  ['stdc++']
if platform.system() == 'Darwin':
    compile_flags += ['--stdlib=libc++']
    cpplib = ['c++']
elif platform.system() == 'Windows':
    compile_flags = ['-DNBUILD', '-DNLGLYALSAT' , '/DINCREMENTAL', '-DNLGLOG',
            '-DNDEBUG', '-DNCHKSOL', '-DNLGLFILES', '-DNLGLDEMA',
            '/experimental:preprocessor', '-I./zlib']
    cpplib = []


# C extensions: pycard and pysolvers
#==============================================================================
pycard_ext = Extension('pycard',
    sources=['cardenc/pycard.cc'],
    extra_compile_args=compile_flags,
    include_dirs=['cardenc'] ,
    language='c++',
    libraries=cpplib,
    library_dirs=[]
)

pysolvers_sources = ['solvers/pysolvers.cc']

if platform.system() == 'Windows':
    with chdir('solvers'):
        for solver in to_install:
            with chdir(solver):
                for filename in glob.glob('*.c*'):
                    pysolvers_sources += ['solvers/%s/%s' % (solver, filename)]
                for filename in glob.glob('*/*.c*'):
                    pysolvers_sources += ['solvers/%s/%s' % (solver, filename)]
    libraries = []
    library_dirs = []
else:
    libraries = to_install + cpplib
    library_dirs = list(map(lambda x: os.path.join('solvers', x), to_install))

pysolvers_ext = Extension('pysolvers',
    sources=pysolvers_sources,
    extra_compile_args=compile_flags + \
        list(map(lambda x: '-DWITH_{0}'.format(x.upper()), to_install)),
    include_dirs=['solvers'],
    language='c++',
    libraries=libraries,
    library_dirs=library_dirs
)


# finally, calling standard setuptools.setup() (or distutils.core.setup())
#==============================================================================
setup(name='python-sat',
    packages=['pysat', 'pysat.examples'],
    package_dir={'pysat.examples': 'examples'},
    version=__version__,
    description='A Python library for prototyping with SAT oracles',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst; charset=UTF-8',
    license='MIT',
    author='Alexey Ignatiev, Joao Marques-Silva, Antonio Morgado',
    author_email='aignatiev@ciencias.ulisboa.pt, jpms@ciencias.ulisboa.pt, ajmorgado@ciencias.ulisboa.pt',
    url='https://github.com/pysathq/pysat',
    ext_modules=[pycard_ext, pysolvers_ext],
    scripts=['examples/{0}.py'.format(s) for s in scripts],
    cmdclass={'build': build},
    install_requires=['six'],
    extras_require = {
        'aiger': ['py-aiger-cnf>=2.0.0'],
        'pblib': ['pypblib>=0.0.3']
    }
)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## models.py
##
##  Created on: Mar 4, 2018
##      Author: Alexey S. Ignatiev
##      E-mail: aignatiev@ciencias.ulisboa.pt
##

"""
    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        enumerate_models

    ==================
    Module description
    ==================

    The module implements a simple iterative enumeration of a given number of
    models of :class:`.CNF` or :class:`CNFPlus` formula. In the latter case,
    only :class:`.Minicard` can be used as a SAT solver. The module aims at
    illustrating how one can work with model computation and enumeration.

    The implementation facilitates the simplest use of a SAT oracle from the
    *command line*. If one deals with the enumeration task from a Python
    script, it is more convenient to exploit the internal model enumeration of
    the :mod:`pysat.solvers` module. Concretely, see
    :meth:`pysat.solvers.Solver.enum_models()`.

    ::

        $ cat formula.cnf
        p cnf 4 4
        -1 2 0
        -1 3 0
        -2 4 0
        3 -4 0

        $ models.py -e all -s glucose3 formula.cnf
        v -1 -2 +3 -4 0
        v +1 +2 -3 +4 0
        c nof models: 2
        c accum time: 0.00s
        c mean  time: 0.00s

    ==============
    Module details
    ==============
"""

#
#==============================================================================
from __future__ import print_function
import getopt
import os
from pysat.formula import CNFPlus
from pysat.solvers import Solver, SolverNames
import sys


#
#==============================================================================
def enumerate_models(formula, to_enum, solver):
    """

        Enumeration procedure. It represents a loop iterating over satisfying
        assignment for a given formula until either all or a given number of
        them is enumerated.

        :param formula: input WCNF formula
        :param to_enum: number of models to compute
        :param solver: name of SAT solver

        :type formula: :class:`.CNFPlus`
        :type to_enum: int or 'all'
        :type solver: str
    """

    with Solver(name=solver, bootstrap_with=formula.clauses, use_timer=True) as s:
        # adding native cardinality constraints if needed
        if formula.atmosts:
            assert solver in SolverNames.minicard, \
                    'Only Minicard supports native cardinality constraints. Make sure you use the right type of formula.'

            for atm in formula.atmosts:
                s.add_atmost(*atm)

        # model enumeration and printing is done here
        computed = 0
        for i, model in enumerate(s.enum_models(), 1):
            print('v {0} 0'.format(' '.join(['{0}{1}'.format('+' if v > 0 else '', v) for v in model])))

            computed = i
            if i == to_enum:
                break

        # some final statistics
        print('c nof models: {0}'.format(computed))
        print('c accum time: {0:.2f}s'.format(s.time()))

        if computed:
            print('c mean  time: {0:.2f}s'.format(s.time() / computed))


#
#==============================================================================
def parse_options():
    """
        Parses command-line options:
    """

    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   'e:h:s:',
                                   ['enum=',
                                    'help',
                                    'solver='])
    except getopt.GetoptError as err:
        sys.stderr.write(str(err).capitalize())
        usage()
        sys.exit(1)

    to_enum = 1
    solver = 'g3'

    for opt, arg in opts:
        if opt in ('-e', '--enum'):
            to_enum = str(arg)
            if to_enum == 'all':
                to_enum = -1
            else:
                to_enum = int(to_enum)
        elif opt in ('-h', '--help'):
            usage()
            sys.exit(0)
        elif opt in ('-s', '--solver'):
            solver = str(arg)
        else:
            assert False, 'Unhandled option: {0} {1}'.format(opt, arg)

    return to_enum, solver, args


#
#==============================================================================
def usage():
    """
        Prints usage message.
    """

    print('Usage:', os.path.basename(sys.argv[0]), '[options] dimacs-file')
    print('Options:')
    print('        -e, --enum=<int>         Compute at most this number of models')
    print('                                 Available values: [1 .. INT_MAX], all (default: 1)')
    print('        -h, --help               Show this message')
    print('        -s, --solver=<string>    SAT solver to use')
    print('                                 Available values: cd, g3, g4, lgl, mcb, mcm, mpl, m22, mc, mgh (default = g3)')


#
#==============================================================================
if __name__ == '__main__':
    # parsing command-line options
    to_enum, solver, files = parse_options()

    # reading an input formula either from a file or from stdin
    if files:
        formula = CNFPlus(from_file=files[0])
    else:
        formula = CNFPlus(from_fp=sys.stdin)

    enumerate_models(formula, to_enum, solver)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## lbx.py
##
##  Created on: Jan 9, 2017
##      Author: Alexey S. Ignatiev
##      E-mail: aignatiev@ciencias.ulisboa.pt
##

"""
    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        LBX

    ==================
    Module description
    ==================

    This module implements a prototype of the LBX algorithm for the computation
    of a *minimal correction subset* (MCS) and/or MCS enumeration. The LBX
    abbreviation stands for *literal-based MCS extraction* algorithm, which was
    proposed in [1]_. Note that this prototype does not follow the original
    low-level implementation of the corresponding MCS extractor available
    `online <https://reason.di.fc.ul.pt/wiki/doku.php?id=lbx>`_ (compared to
    our prototype, the low-level implementation has a number of additional
    heuristics used). However, it implements the LBX algorithm for partial
    MaxSAT formulas, as described in [1]_.

    .. [1] Carlos Mencia, Alessandro Previti, Joao Marques-Silva.
        *Literal-Based MCS Extraction*. IJCAI 2015. pp. 1973-1979

    The implementation can be used as an executable (the list of available
    command-line options can be shown using ``lbx.py -h``) in the following
    way:

    ::

        $ xzcat formula.wcnf.xz
        p wcnf 3 6 4
        1 1 0
        1 2 0
        1 3 0
        4 -1 -2 0
        4 -1 -3 0
        4 -2 -3 0

        $ lbx.py -d -e all -s glucose3 -vv formula.wcnf.xz
        c MCS: 1 3 0
        c cost: 2
        c MCS: 2 3 0
        c cost: 2
        c MCS: 1 2 0
        c cost: 2
        c oracle time: 0.0002

    Alternatively, the algorithm can be accessed and invoked through the
    standard ``import`` interface of Python, e.g.

    .. code-block:: python

        >>> from pysat.examples.lbx import LBX
        >>> from pysat.formula import WCNF
        >>>
        >>> wcnf = WCNF(from_file='formula.wcnf.xz')
        >>>
        >>> lbx = LBX(wcnf, use_cld=True, solver_name='g3')
        >>> for mcs in lbx.enumerate():
        ...     lbx.block(mcs)
        ...     print(mcs)
        [1, 3]
        [2, 3]
        [1, 2]

    ==============
    Module details
    ==============
"""

#
#==============================================================================
from __future__ import print_function
import collections
import getopt
from math import copysign
import os
from pysat.formula import CNFPlus, WCNFPlus
from pysat.solvers import Solver, SolverNames
import re
from six.moves import range
import sys


#
#==============================================================================
class LBX(object):
    """
        LBX-like algorithm for computing MCSes. Given an unsatisfiable partial
        CNF formula, i.e. formula in the :class:`.WCNF` format, this class can
        be used to compute a given number of MCSes of the formula. The
        implementation follows the LBX algorithm description in [1]_. It can
        use any SAT solver available in PySAT. Additionally, the "clause
        :math:`D`" heuristic can be used when enumerating MCSes.

        The default SAT solver to use is ``m22`` (see :class:`.SolverNames`).
        The "clause :math:`D`" heuristic is disabled by default, i.e.
        ``use_cld`` is set to ``False``. Internal SAT solver's timer is also
        disabled by default, i.e. ``use_timer`` is ``False``.

        :param formula: unsatisfiable partial CNF formula
        :param use_cld: whether or not to use "clause :math:`D`"
        :param solver_name: SAT oracle name
        :param use_timer: whether or not to use SAT solver's timer

        :type formula: :class:`.WCNF`
        :type use_cld: bool
        :type solver_name: str
        :type use_timer: bool
    """

    def __init__(self, formula, use_cld=False, solver_name='m22', use_timer=False):
        """
            Constructor.
        """

        # bootstrapping the solver with hard clauses
        self.oracle = Solver(name=solver_name, bootstrap_with=formula.hard,
                use_timer=use_timer)
        self.solver = solver_name

        # adding native cardinality constraints (if any) as hard clauses
        # this can be done only if the Minicard solver is in use
        if isinstance(formula, WCNFPlus) and formula.atms:
            assert solver_name in SolverNames.minicard, \
                    'Only Minicard supports native cardinality constraints. Make sure you use the right type of formula.'

            for atm in formula.atms:
                self.oracle.add_atmost(*atm)

        self.topv = formula.nv  # top variable id
        self.soft = formula.soft
        self.sels = []
        self.ucld = use_cld

        # mappings between internal and external variables
        VariableMap = collections.namedtuple('VariableMap', ['e2i', 'i2e'])
        self.vmap = VariableMap(e2i={}, i2e={})

        # at this point internal and external variables are the same
        for v in range(1, formula.nv + 1):
            self.vmap.e2i[v] = v
            self.vmap.i2e[v] = v

        for cl in self.soft:
            sel = cl[0]
            if len(cl) > 1 or cl[0] < 0:
                self.topv += 1
                sel = self.topv

                self.oracle.add_clause(cl + [-sel])

            self.sels.append(sel)

    def __del__(self):
        """
            Destructor.
        """

        self.delete()

    def __enter__(self):
        """
            'with' constructor.
        """

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
            'with' destructor.
        """

        self.delete()

    def delete(self):
        """
            Explicit destructor of the internal SAT oracle.
        """

        if self.oracle:
            self.oracle.delete()
            self.oracle = None

    def add_clause(self, clause, soft=False):
        """
            The method for adding a new hard of soft clause to the problem
            formula. Although the input formula is to be specified as an
            argument of the constructor of :class:`LBX`, adding clauses may be
            helpful when *enumerating* MCSes of the formula. This way, the
            clauses are added incrementally, i.e. *on the fly*.

            The clause to add can be any iterable over integer literals. The
            additional Boolean parameter ``soft`` can be set to ``True``
            meaning the the clause being added is soft (note that parameter
            ``soft`` is set to ``False`` by default).

            :param clause: a clause to add
            :param soft: whether or not the clause is soft

            :type clause: iterable(int)
            :type soft: bool
        """

        # first, map external literals to internal literals
        # introduce new variables if necessary
        cl = list(map(lambda l: self._map_extlit(l), clause if not len(clause) == 2 or not type(clause[0]) == list else clause[0]))

        if not soft:
            if not len(clause) == 2 or not type(clause[0]) == list:
                # the clause is hard, and so we simply add it to the SAT oracle
                self.oracle.add_clause(cl)
            else:
                # this should be a native cardinality constraint,
                # which can be used only together with Minicard
                assert self.solver in SolverNames.minicard, \
                        'Only Minicard supports native cardinality constraints.'

                self.oracle.add_atmost(cl, clause[1])
        else:
            self.soft.append(cl)

            # soft clauses should be augmented with a selector
            sel = cl[0]
            if len(cl) > 1 or cl[0] < 0:
                self.topv += 1
                sel = self.topv

                self.oracle.add_clause(cl + [-sel])

            self.sels.append(sel)

    def compute(self):
        """
            Compute and return one solution. This method checks whether the
            hard part of the formula is satisfiable, i.e. an MCS can be
            extracted. If the formula is satisfiable, the model computed by the
            SAT call is used as an *over-approximation* of the MCS in the
            method :func:`_compute` invoked here, which implements the LBX
            algorithm.

            An MCS is reported as a list of integers, each representing a soft
            clause index (the smallest index is ``1``).

            :rtype: list(int)
        """

        self.setd = []
        self.satc = [False for cl in self.soft]  # satisfied clauses
        self.solution = None
        self.bb_assumps = []  # backbone assumptions
        self.ss_assumps = []  # satisfied soft clause assumptions

        if self.oracle.solve():
            # hard part is satisfiable => there is a solution
            self._filter_satisfied(update_setd=True)
            self._compute()

            self.solution = list(map(lambda i: i + 1, filter(lambda i: not self.satc[i], range(len(self.soft)))))

        return self.solution

    def enumerate(self):
        """
            This method iterates through MCSes enumerating them until the
            formula has no more MCSes. The method iteratively invokes
            :func:`compute`. Note that the method does not block the MCSes
            computed - this should be explicitly done by a user.
        """

        done = False
        while not done:
            mcs = self.compute()

            if mcs != None:
                yield mcs
            else:
                done = True

    def block(self, mcs):
        """
            Block a (previously computed) MCS. The MCS should be given as an
            iterable of integers. Note that this method is not automatically
            invoked from :func:`enumerate` because a user may want to block
            some of the MCSes conditionally depending on the needs. For
            example, one may want to compute disjoint MCSes only in which case
            this standard blocking is not appropriate.

            :param mcs: an MCS to block
            :type mcs: iterable(int)
        """

        self.oracle.add_clause([self.sels[cl_id - 1] for cl_id in mcs])

    def _satisfied(self, cl, model):
        """
            Given a clause (as an iterable of integers) and an assignment (as a
            list of integers), this method checks whether or not the assignment
            satisfies the clause. This is done by a simple clause traversal.
            The method is invoked from :func:`_filter_satisfied`.

            :param cl: a clause to check
            :param model: an assignment

            :type cl: iterable(int)
            :type model: list(int)

            :rtype: bool
        """

        for l in cl:
            if len(model) < abs(l) or model[abs(l) - 1] == l:
                # either literal is unassigned or satisfied by the model
                return True

        return False

    def _filter_satisfied(self, update_setd=False):
        """
            This method extracts a model provided by the previous call to a SAT
            oracle and iterates over all soft clauses checking if each of is
            satisfied by the model. Satisfied clauses are marked accordingly
            while the literals of the unsatisfied clauses are kept in a list
            called ``setd``, which is then used to refine the correction set
            (see :func:`_compute`, and :func:`do_cld_check`).

            Optional Boolean parameter ``update_setd`` enforces the method to
            update variable ``self.setd``. If this parameter is set to
            ``False``, the method only updates the list of satisfied clauses,
            which is an under-approximation of a *maximal satisfiable subset*
            (MSS).

            :param update_setd: whether or not to update setd
            :type update_setd: bool
        """

        model = self.oracle.get_model()
        setd = set()

        for i, cl in enumerate(self.soft):
            if not self.satc[i]:
                if self._satisfied(cl, model):
                    self.satc[i] = True
                    self.ss_assumps.append(self.sels[i])
                else:
                    setd = setd.union(set(cl))

        if update_setd:
            self.setd = list(setd)

    def _compute(self):
        """
            The main method of the class, which computes an MCS given its
            over-approximation. The over-approximation is defined by a model
            for the hard part of the formula obtained in :func:`compute`.

            The method is essentially a simple loop going over all literals
            unsatisfied by the previous model, i.e. the literals of
            ``self.setd`` and checking which literals can be satisfied. This
            process can be seen a refinement of the over-approximation of the
            MCS. The algorithm follows the pseudo-code of the LBX algorithm
            presented in [1]_.

            Additionally, if :class:`LBX` was constructed with the requirement
            to make "clause :math:`D`" calls, the method calls
            :func:`do_cld_check` at every iteration of the loop using the
            literals of ``self.setd`` not yet checked, as the contents of
            "clause :math:`D`".
        """

        # unless clause D checks are used, test one literal at a time
        # and add it either to satisfied of backbone assumptions
        i = 0
        while i < len(self.setd):
            if self.ucld:
                self.do_cld_check(self.setd[i:])
                i = 0

            if self.setd:  # if may be empty after the clause D check
                if self.oracle.solve(assumptions=self.ss_assumps + self.bb_assumps + [self.setd[i]]):
                    # filtering satisfied clauses
                    self._filter_satisfied()
                else:
                    # current literal is backbone
                    self.bb_assumps.append(-self.setd[i])

            i += 1

    def do_cld_check(self, cld):
        """
            Do the "clause :math:`D`" check. This method receives a list of
            literals, which serves a "clause :math:`D`" [2]_, and checks
            whether the formula conjoined with :math:`D` is satisfiable.

            .. [2] Joao Marques-Silva, Federico Heras, Mikolas Janota,
                Alessandro Previti, Anton Belov. *On Computing Minimal
                Correction Subsets*. IJCAI 2013. pp. 615-622

            If clause :math:`D` cannot be satisfied together with the formula,
            then negations of all of its literals are backbones of the formula
            and the LBX algorithm can stop. Otherwise, the literals satisfied
            by the new model refine the MCS further.

            Every time the method is called, a new fresh selector variable
            :math:`s` is introduced, which augments the current clause
            :math:`D`. The SAT oracle then checks if clause :math:`(D \\vee
            \\neg{s})` can be satisfied together with the internal formula.
            The :math:`D` clause is then disabled by adding a hard clause
            :math:`(\\neg{s})`.

            :param cld: clause :math:`D` to check
            :type cld: list(int)
        """

        # adding a selector literal to clause D
        # selector literals for clauses D currently
        # cannot be reused, but this may change later
        self.topv += 1
        sel = self.topv
        cld.append(-sel)

        # adding clause D
        self.oracle.add_clause(cld)

        if self.oracle.solve(assumptions=self.ss_assumps + self.bb_assumps + [sel]):
            # filtering satisfied
            self._filter_satisfied(update_setd=True)
        else:
            # clause D is unsatisfiable => all literals are backbones
            self.bb_assumps.extend([-l for l in cld[:-1]])
            self.setd = []

        # deactivating clause D
        self.oracle.add_clause([-sel])

    def _map_extlit(self, l):
        """
            Map an external variable to an internal one if necessary.

            This method is used when new clauses are added to the formula
            incrementally, which may result in introducing new variables
            clashing with the previously used *clause selectors*. The method
            makes sure no clash occurs, i.e. it maps the original variables
            used in the new problem clauses to the newly introduced auxiliary
            variables (see :func:`add_clause`).

            Given an integer literal, a fresh literal is returned. The returned
            integer has the same sign as the input literal.

            :param l: literal to map
            :type l: int

            :rtype: int
        """

        v = abs(l)

        if v in self.vmap.e2i:
            return int(copysign(self.vmap.e2i[v], l))
        else:
            self.topv += 1

            self.vmap.e2i[v] = self.topv
            self.vmap.i2e[self.topv] = v

            return int(copysign(self.topv, l))

    def oracle_time(self):
        """
            Report the total SAT solving time.
        """

        return self.oracle.time_accum()


#
#==============================================================================
def parse_options():
    """
        Parses command-line options.
    """

    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   'de:hs:v',
                                   ['dcalls',
                                    'enum=',
                                    'help',
                                    'solver=',
                                    'verbose'])
    except getopt.GetoptError as err:
        sys.stderr.write(str(err).capitalize() + '\n')
        usage()
        sys.exit(1)

    dcalls = False
    to_enum = 1
    solver = 'm22'
    verbose = 0

    for opt, arg in opts:
        if opt in ('-d', '--dcalls'):
            dcalls = True
        elif opt in ('-e', '--enum'):
            to_enum = str(arg)
            if to_enum != 'all':
                to_enum = int(to_enum)
        elif opt in ('-h', '--help'):
            usage()
            sys.exit(0)
        elif opt in ('-s', '--solver'):
            solver = str(arg)
        elif opt in ('-v', '--verbose'):
            verbose += 1
        else:
            assert False, 'Unhandled option: {0} {1}'.format(opt, arg)

    return dcalls, to_enum, solver, verbose, args


#
#==============================================================================
def usage():
    """
        Prints help message.
    """

    print('Usage:', os.path.basename(sys.argv[0]), '[options] file')
    print('Options:')
    print('        -d, --dcalls           Try to bootstrap algorithm')
    print('        -e, --enum=<string>    How many solutions to compute')
    print('                               Available values: [1 .. all] (default: 1)')
    print('        -h, --help')
    print('        -s, --solver           SAT solver to use')
    print('                               Available values: g3, g4, lgl, mcb, mcm, mpl, m22, mc, mgh (default = m22)')
    print('        -v, --verbose          Be verbose')


#
#==============================================================================
if __name__ == '__main__':
    dcalls, to_enum, solver, verbose, files = parse_options()

    if type(to_enum) == str:
        to_enum = 0

    if files:
        # reading standard CNF, WCNF, or (W)CNF+
        if re.search('cnf[p|+]?(\.(gz|bz2|lzma|xz))?$', files[0]):
            if re.search('\.wcnf[p|+]?(\.(gz|bz2|lzma|xz))?$', files[0]):
                formula = WCNFPlus(from_file=files[0])
            else:  # expecting '*.cnf[,p,+].*'
                formula = CNFPlus(from_file=files[0]).weighted()

        with LBX(formula, use_cld=dcalls, solver_name=solver, use_timer=True) as mcsls:
            for i, mcs in enumerate(mcsls.enumerate()):
                if verbose:
                    print('c MCS:', ' '.join([str(cl_id) for cl_id in mcs]), '0')

                    if verbose > 1:
                        cost = sum([formula.wght[cl_id - 1] for cl_id in mcs])
                        print('c cost:', cost)

                if to_enum and i + 1 == to_enum:
                    break

                mcsls.block(mcs)

            print('c oracle time: {0:.4f}'.format(mcsls.oracle_time()))
